import atexit
import pathlib
import shlex
import shutil
import sys
import tempfile
from typing import Annotated, List, Optional, Union

import rich
import rich.prompt
import syncer
import typer
from ordered_set import OrderedSet

from rbx import annotations, config, console, utils
from rbx.annotations import PackagePath
from rbx.box import (
    cd,
    compile,
    creation,
    download,
    environment,
    generator_script_handlers,
    generators,
    global_package,
    limits_info,
    package,
    presets,
    setter_config,
    state,
    summary,
    timing,
    validators,
)
from rbx.box.contest import main as contest
from rbx.box.contest.contest_package import find_contest_yaml
from rbx.box.environment import VerificationLevel, get_app_environment_path
from rbx.box.generation_schema import get_parsed_entry
from rbx.box.header import generate_header
from rbx.box.packaging import main as packaging
from rbx.box.schema import ExpectedOutcome, GeneratorScript, TestcaseGroup
from rbx.box.solutions import (
    get_exact_matching_solutions,
    get_matching_solutions,
    pick_solutions,
    print_run_report,
    run_and_print_interactive_solutions,
    run_solutions,
)
from rbx.box.statements import build_statements
from rbx.box.testcases import main as testcases
from rbx.box.tooling import main as tooling
from rbx.grading import grading_context

app = typer.Typer(
    no_args_is_help=True, add_completion=False, cls=annotations.AliasGroup
)
app.add_typer(
    setter_config.app,
    name='config, cfg',
    cls=annotations.AliasGroup,
    help='Manage setter configuration (sub-command).',
    rich_help_panel='Configuration',
)
app.add_typer(
    build_statements.app,
    name='statements, st',
    cls=annotations.AliasGroup,
    help='Manage statements (sub-command).',
    rich_help_panel='Deploying',
)
app.add_typer(
    download.app,
    name='download, down',
    cls=annotations.AliasGroup,
    help='Download an asset from supported repositories (sub-command).',
    rich_help_panel='Management',
)
app.add_typer(
    presets.app,
    name='presets',
    cls=annotations.AliasGroup,
    help='Manage presets (sub-command).',
    rich_help_panel='Configuration',
)
app.add_typer(
    packaging.app,
    name='package, pkg',
    cls=annotations.AliasGroup,
    help='Build problem packages (sub-command).',
    rich_help_panel='Deploying',
)
app.add_typer(
    contest.app,
    name='contest',
    cls=annotations.AliasGroup,
    help='Manage contests (sub-command).',
    rich_help_panel='Management',
)
app.add_typer(
    testcases.app,
    name='testcases, tc, t',
    cls=annotations.AliasGroup,
    help='Manage testcases (sub-command).',
    rich_help_panel='Management',
)
app.add_typer(
    tooling.app,
    name='tool, tooling',
    cls=annotations.AliasGroup,
    help='Manage tooling (sub-command).',
    rich_help_panel='Misc',
)


def version_callback(value: bool) -> None:
    if value:
        version = utils.get_version()

        console.console.print(f'rbx version {version}')
        raise typer.Exit()


@app.callback()
def main(
    cache: Annotated[
        grading_context.CacheLevel,
        typer.Option(
            '-c',
            '--cache',
            help='Which degree of caching to use.',
            default_factory=lambda: grading_context.CacheLevel.CACHE_ALL.value,
        ),
    ],
    sanitized: bool = typer.Option(
        False,
        '--sanitized',
        '-s',
        help='Whether to compile and run testlib components with sanitizers enabled. '
        'If you want to run the solutions with sanitizers enabled, use the "-s" flag in the corresponding run command.',
    ),
    capture: bool = typer.Option(
        True,
        '--nocapture',
        help='Whether to save extra logs and outputs from interactive solutions.',
    ),
    profile: Annotated[
        Optional[str],
        typer.Option(
            '-p',
            '--profile',
            help='Which timing profile to use when running solutions.',
        ),
    ] = None,
    profiling: bool = typer.Option(
        False,
        '--profiling',
        help='Whether to profile (capture performance statistics) of the execution.',
    ),
    version: Annotated[
        bool, typer.Option('--version', '-v', callback=version_callback, is_eager=True)
    ] = False,
):
    # Load .env variables.
    utils.load_dotenv()

    presets.check_active_preset_compatibility()
    if cd.is_problem_package() and not package.is_cache_valid():
        console.console.print(
            '[warning]Cache is incompatible with the current version of [item]rbx[/item], so it will be cleared.[/warning]'
        )
        clear()
    if not global_package.is_global_cache_valid():
        console.console.print(
            '[warning]Global cache is incompatible with the current version of [item]rbx[/item], so it will be cleared.[/warning]'
        )
        clear(global_cache=True)

    state.STATE.run_through_cli = True
    state.STATE.sanitized = sanitized
    if sanitized:
        console.console.print(
            '[warning]Sanitizers are running just for testlib components.\n'
            'If you want to run the solutions with sanitizers enabled, use the [item]-s[/item] flag in the corresponding run command.[/warning]'
        )
    state.STATE.debug_logs = capture

    grading_context.cache_level_var.set(grading_context.CacheLevel(cache))
    grading_context.check_integrity_var.set(
        setter_config.get_setter_config().caching.check_integrity
    )

    if profile is not None:
        from rbx.box import limits_info

        limits_info.profile_var.set(profile)

    if profiling:
        from rbx.grading import profiling as profiling_module

        atexit.register(profiling_module.print_summary)


@app.command('ui', help='Show an UI for exploring testcases of the current problem.')
@package.within_problem
def ui():
    from rbx.box.ui import main as ui_pkg

    ui_pkg.start()


@app.command(
    'on',
    help='Run a command in the context of a problem (or a set of problems) of a contest.',
    context_settings={'allow_extra_args': True, 'ignore_unknown_options': True},
)
def on(ctx: typer.Context, problems: str) -> None:
    contest.on(ctx, problems)


@app.command(
    'each',
    help='Run a command for each problem in the contest.',
    context_settings={'allow_extra_args': True, 'ignore_unknown_options': True},
)
def each(ctx: typer.Context) -> None:
    contest.each(ctx)


@app.command('diff', hidden=True)
def diff(path1: pathlib.Path, path2: pathlib.Path):
    from rbx.box.ui import main as ui_pkg

    ui_pkg.start_differ(path1, path2)


@app.command('serve', hidden=True)
def serve():
    from textual_serve.server import Server

    server = Server('rbx ui', port=8081)
    server.serve()


@app.command(
    'edit, e',
    rich_help_panel='Configuration',
    help='Open problem.rbx.yml in your default editor.',
)
@package.within_problem
def edit():
    console.console.print('Opening problem definition in editor...')
    # Call this function just to raise exception in case we're no in
    # a problem package.
    package.find_problem()
    config.open_editor(package.find_problem_yaml() or pathlib.Path())


@app.command(
    'build, b', rich_help_panel='Deploying', help='Build all tests for the problem.'
)
@package.within_problem
@syncer.sync
async def build(
    verification: environment.VerificationParam,
    validate: bool = typer.Option(
        True,
        help='Whether to validate outputs for tests.',
    ),
    visualize: bool = typer.Option(
        False,
        help='Whether to build visualizations for inputs/outputs of tests.',
    ),
):
    from rbx.box import builder

    await builder.build(
        verification=verification, validate=validate, visualize=visualize
    )


@app.command(
    'run, r',
    rich_help_panel='Testing',
    help='Build and run solution(s).',
)
@package.within_problem
@syncer.sync
async def run(
    verification: environment.VerificationParam,
    solutions: Annotated[
        Optional[List[str]],
        PackagePath,
        typer.Argument(
            help='Path to solutions to run. If not specified, will run all solutions.'
        ),
    ] = None,
    outcome: Optional[str] = typer.Option(
        None,
        '--outcome',
        '-o',
        help='Include only solutions whose expected outcomes intersect with this.',
    ),
    tags: Annotated[
        Optional[List[str]],
        typer.Option(
            '--tag',
            '-t',
            help='Include only solutions whose tags intersect with this.',
        ),
    ] = None,
    check: bool = typer.Option(
        True,
        help='Whether to not build outputs for tests and run checker.',
    ),
    validate: bool = typer.Option(
        True,
        help='Whether to not validate outputs for tests.',
    ),
    detailed: bool = typer.Option(
        False,
        '--detailed',
        '-d',
        help='Whether to print a detailed view of the tests using tables.',
    ),
    sanitized: bool = typer.Option(
        False,
        '--sanitized',
        '-s',
        help='Whether to compile the solutions with sanitizers enabled.',
    ),
    choice: bool = typer.Option(
        False,
        '--choice',
        '--choose',
        '-c',
        help='Whether to pick solutions interactively.',
    ),
):
    main_solution = package.get_main_solution()
    if check and main_solution is None:
        console.console.print(
            '[warning]No main solution found, running without checkers.[/warning]'
        )
        check = False

    tracked_solutions: Optional[OrderedSet[str]] = None
    if outcome is not None or tags is not None:
        tracked_solutions = OrderedSet(
            str(solution.path)
            for solution in get_matching_solutions(
                expected_outcome=ExpectedOutcome(outcome)
                if outcome is not None
                else None,
                tags=tags or None,
            )
        )
    if solutions:
        tracked_solutions = OrderedSet(solutions)

    if choice:
        tracked_solutions = OrderedSet(
            await pick_solutions(
                tracked_solutions,
                extra_solutions=solutions,
            )
        )
        if not tracked_solutions:
            console.console.print('[error]No solutions selected. Exiting.[/error]')
            raise typer.Exit(1)

    from rbx.box import builder

    if not await builder.build(
        verification=verification, output=check, validate=validate, is_run=True
    ):
        return

    if verification <= VerificationLevel.VALIDATE.value:
        console.console.print(
            '[warning]Verification level is set to [item]validate (-v1)[/item], so rbx only build tests and validated them.[/warning]'
        )
        console.console.print(
            '[warning]If you want to run solutions, but skip validation, run with [item]--no-validate[/item].[/warning]'
        )
        return

    if sanitized:
        console.console.print(
            '[warning]Sanitizers are running, so the time limit for the problem will be dropped, '
            'and the environment default time limit will be used instead.[/warning]'
        )

    if sanitized and tracked_solutions is None:
        console.console.print(
            '[warning]Sanitizers are running, and no solutions were specified to run. Will only run [item]ACCEPTED[/item] solutions.'
        )
        tracked_solutions = OrderedSet(
            str(solution.path)
            for solution in get_exact_matching_solutions(ExpectedOutcome.ACCEPTED)
        )

    with utils.StatusProgress('Running solutions...') as s:
        solution_result = run_solutions(
            progress=s,
            tracked_solutions=tracked_solutions,
            check=check,
            verification=VerificationLevel(verification),
            sanitized=sanitized,
        )

    console.console.print()
    console.console.rule('[status]Run report[/status]', style='status')
    await print_run_report(
        solution_result,
        console.console,
        VerificationLevel(verification),
        detailed=detailed,
        skip_printing_limits=sanitized,
    )


@app.command(
    'summary, sum',
    rich_help_panel='Testing',
    help='Print a summary of the problem.',
)
@package.within_problem
@syncer.sync
async def summary_cmd(
    detailed: bool = typer.Option(
        False,
        '--detailed',
        '-d',
        help='Whether to print a detailed view of the tests using tables.',
    ),
):
    await summary.print_problem_summary(
        package.find_problem_package_or_die(), detailed=detailed
    )


@app.command(
    'time, t',
    rich_help_panel='Testing',
    help='Estimate a time limit for the problem based on a time limit formula and timings of accepted solutions.',
)
@package.within_problem
@syncer.sync
async def time(
    check: bool = typer.Option(
        True,
        help='Whether to not build outputs for tests and run checker.',
    ),
    validate: bool = typer.Option(
        True,
        help='Whether to not validate outputs for tests.',
    ),
    detailed: bool = typer.Option(
        False,
        '--detailed',
        '-d',
        help='Whether to print a detailed view of the tests using tables.',
    ),
    strategy: Optional[str] = typer.Option(
        None,
        '--strategy',
        '-s',
        help='Strategy to use for time limit estimation (estimate, inherit).',
    ),
    auto: bool = typer.Option(
        False,
        '--auto',
        '-a',
        help='Whether to automatically estimate the time limit.',
    ),
    runs: int = typer.Option(
        0,
        '--runs',
        '-r',
        help='Number of runs to perform for each solution. Zero means the config default.',
    ),
    profile: str = typer.Option(
        'local',
        '--profile',
        '-p',
        help='Profile to use for time limit estimation.',
    ),
    integrate: bool = typer.Option(
        False,
        '--integrate',
        '-i',
        help='Integrate the given limits profile into the package.',
    ),
):
    current_profile = limits_info.get_limits_profile(profile)
    if current_profile is not None:
        console.console.print(
            f'[bright_white]Current limits for profile [item]{profile}[/item]:[/bright_white]'
        )
        limits_info.pretty_print_profile(current_profile)
        console.console.print()
    if integrate:
        timing.integrate(profile)
        return

    if auto:
        strategy = 'estimate'

    import questionary

    formula = environment.get_environment().timing.formula
    timing_choices = [
        questionary.Choice(
            f'Estimate time limits based on the formula {formula} (recommended)',
            value='estimate',
        ),
        questionary.Choice('Inherit from the package.', value='inherit'),
        questionary.Choice(
            'Estimate time limits based on a custom formula.', value='estimate_custom'
        ),
        questionary.Choice('Provide a custom time limit.', value='custom'),
    ]

    choice = (
        strategy
        or await questionary.select(
            'Select how you want to define the time limits for the problem.',
            choices=timing_choices,
        ).ask_async()
    )
    if choice is None:
        console.console.print(
            '[error]No time limit strategy selected. Exiting.[/error]'
        )
        raise typer.Exit(1)

    formula = environment.get_environment().timing.formula

    if choice == 'inherit':
        timing.inherit_time_limits(profile=profile)
        return
    elif choice == 'custom':
        timelimit = await questionary.text(
            'Enter a custom time limit for the problem (ms).',
            validate=lambda x: x.isdigit() and int(x) > 0,
        ).ask_async()
        if timelimit is None:
            console.console.print(
                '[error]No custom time limit provided. Exiting.[/error]'
            )
            raise typer.Exit(1)
        timing.set_time_limit(int(timelimit), profile=profile)
        return

    if choice == 'estimate_custom':
        formula = await questionary.text(
            'Enter a custom formula for time limit estimation.'
        ).ask_async()
        if formula is None:
            console.console.print('[error]No custom formula provided. Exiting.[/error]')
            raise typer.Exit(1)
    main_solution = package.get_main_solution()
    if check and main_solution is None:
        console.console.print(
            '[warning]No main solution found, running without checkers.[/warning]'
        )
        check = False

    from rbx.box import builder

    verification = VerificationLevel.ALL_SOLUTIONS.value
    if not await builder.build(
        verification=verification, output=check, validate=validate, is_run=True
    ):
        return None

    await timing.compute_time_limits(
        check, detailed, runs, formula=formula, profile=profile, auto=auto
    )


@app.command(
    'irun, ir',
    rich_help_panel='Testing',
    help='Build and run solution(s) by passing testcases in the CLI.',
)
@package.within_problem
@syncer.sync
async def irun(
    verification: environment.VerificationParam,
    solutions: Annotated[
        Optional[List[str]],
        PackagePath,
        typer.Argument(
            help='Path to solutions to run. If not specified, will run all solutions.'
        ),
    ] = None,
    outcome: Optional[str] = typer.Option(
        None,
        '--outcome',
        '-o',
        help='Include only solutions whose expected outcomes intersect with this.',
    ),
    tags: Annotated[
        Optional[List[str]],
        typer.Option(
            '--tag',
            '-t',
            help='Include only solutions whose tags intersect with this.',
        ),
    ] = None,
    check: bool = typer.Option(
        True,
        help='Whether to not build outputs for tests and run checker.',
    ),
    validate: bool = typer.Option(
        True,
        help='Whether to validate inputs.',
    ),
    generator: Optional[str] = typer.Option(
        None,
        '--generator',
        '-g',
        help='Generator call to use to generate a single test for execution.',
    ),
    testcase: Optional[str] = typer.Option(
        None,
        '--testcase',
        '--test',
        '-tc',
        '-t',
        help='Testcase to run, in the format "[group]/[index]". If not specified, will run interactively.',
    ),
    output: bool = typer.Option(
        False,
        '--output',
        '-O',
        help='Whether to ask user for custom output.',
    ),
    visualize: bool = typer.Option(
        False,
        '--visualize',
        help='Whether to generate visualizations for inputs and outputs.',
    ),
    print: bool = typer.Option(
        False, '--print', '-p', help='Whether to print outputs to terminal.'
    ),
    sanitized: bool = typer.Option(
        False,
        '--sanitized',
        '-s',
        help='Whether to compile the solutions with sanitizers enabled.',
    ),
    choice: bool = typer.Option(
        False,
        '--choice',
        '--choose',
        '-c',
        help='Whether to pick solutions interactively.',
    ),
):
    if not print:
        console.console.print(
            '[warning]Outputs will be written to files. If you wish to print them to the terminal, use the "-p" parameter.'
        )
    if verification < VerificationLevel.ALL_SOLUTIONS.value:
        console.console.print(
            '[warning]Verification level should be at least [item]all solutions (-v4)[/item] to run solutions interactively.'
        )
        return

    tracked_solutions: Optional[OrderedSet[str]] = None
    if outcome is not None or tags is not None:
        tracked_solutions = OrderedSet(
            str(solution.path)
            for solution in get_matching_solutions(
                expected_outcome=ExpectedOutcome(outcome)
                if outcome is not None
                else None,
                tags=tags or None,
            )
        )
    if solutions:
        tracked_solutions = OrderedSet(solutions)

    if choice:
        tracked_solutions = OrderedSet(
            await pick_solutions(
                tracked_solutions,
                extra_solutions=solutions,
            )
        )
        if not tracked_solutions:
            console.console.print('[error]No solutions selected. Exiting.[/error]')
            raise typer.Exit(1)

    if sanitized and tracked_solutions is None:
        console.console.print(
            '[warning]Sanitizers are running, and no solutions were specified to run. Will only run [item]ACCEPTED[/item] solutions.'
        )
        tracked_solutions = OrderedSet(
            str(solution.path)
            for solution in get_exact_matching_solutions(ExpectedOutcome.ACCEPTED)
        )

    with utils.StatusProgress('Running solutions...') as s:
        await run_and_print_interactive_solutions(
            progress=s,
            tracked_solutions=tracked_solutions,
            check=check,
            verification=VerificationLevel(verification),
            generator=generators.get_call_from_string(generator)
            if generator is not None
            else None,
            testcase_entry=get_parsed_entry(testcase) if testcase else None,
            custom_output=output,
            print=print,
            sanitized=sanitized,
            validate=validate,
            visualize=visualize,
        )


@app.command(
    'create, c',
    rich_help_panel='Management',
    help='Create a new problem package.',
)
def create(
    name: Annotated[
        str,
        typer.Option(
            help='Name of the problem to create, which will be used as the name of the new folder.',
            prompt='What should the name of the problem be?',
        ),
    ],
    preset: Annotated[
        Optional[str], typer.Option(help='Preset to use when creating the problem.')
    ] = None,
):
    if find_contest_yaml() is not None:
        console.console.print(
            '[error]Cannot [item]rbx create[/item] a problem inside a contest.[/error]'
        )
        console.console.print(
            '[error]Instead, use [item]rbx contest add[/item] to add a problem to a contest.[/error]'
        )
        raise typer.Exit(1)

    if preset is not None:
        creation.create(name, preset=preset)
        return
    creation.create(name)


@app.command(
    'stress',
    rich_help_panel='Testing',
    help='Run a stress test.',
)
@package.within_problem
@syncer.sync
async def stress(
    name: Annotated[
        Optional[str],
        typer.Argument(
            help='Name of the stress test to run (specified in problem.rbx.yml).'
        ),
    ] = None,
    generator_args: Annotated[
        Optional[str],
        typer.Option(
            '--generator',
            '-g',
            help='Generator call to use to generate a single test for execution.',
        ),
    ] = None,
    finder: Annotated[
        Optional[str],
        typer.Option(
            '--finder',
            '-f',
            help='Run a stress with this finder expression.',
        ),
    ] = None,
    timeout: Annotated[
        int,
        typer.Option(
            '--timeout',
            '--time',
            '-t',
            help='For how many seconds to run the stress test.',
        ),
    ] = 10,
    findings: Annotated[
        int,
        typer.Option('--findings', '-n', help='How many breaking tests to look for.'),
    ] = 1,
    verbose: bool = typer.Option(
        False,
        '-v',
        '--verbose',
        help='Whether to print verbose output for checkers and finders.',
    ),
    sanitized: bool = typer.Option(
        False,
        '--sanitized',
        '-s',
        help='Whether to compile the solutions with sanitizers enabled.',
    ),
    description: Optional[str] = typer.Option(
        None,
        '--description',
        '-d',
        help='Optional description of the stress test.',
    ),
    print_descriptors: bool = typer.Option(
        False,
        '--descriptors',
        '-D',
        help='Whether to print descriptors of the stress test.',
        hidden=True,
    ),
    skip_invalid_testcases: bool = typer.Option(
        False,
        '--skip-invalid',
        '--skip',
        help='Whether to skip invalid testcases.',
    ),
    custom_timelimit: Annotated[
        Optional[int],
        typer.Option('--timelimit', '-T', help='Custom timelimit for the stress test.'),
    ] = None,
    double_timelimit: Annotated[
        bool,
        typer.Option(
            '--double-tl',
            help='Whether to use 2*TL as the timelimit for the stress test.',
        ),
    ] = False,
    find_slowest: Annotated[
        bool,
        typer.Option(
            '--slowest',
            help='Whether to find the slowest testcases. This removes the time limit of the solution '
            'executions and focus on finding the testcases that make them the slowest.',
        ),
    ] = False,
    fuzz: Annotated[
        bool,
        typer.Option(
            '--fuzz',
            help='Whether to fuzz generator calls from all testgroups.',
        ),
    ] = False,
    fuzz_on: Annotated[
        Optional[List[str]],
        typer.Option(
            '--fuzz-on',
            help='Testgroups to fuzz generator calls from.',
        ),
    ] = None,
    validate: bool = typer.Option(
        True,
        help='Whether to validate inputs.',
    ),
):
    if generator_args and (fuzz or fuzz_on):
        console.console.print(
            '[error]Options --generator/-g and --fuzz/--fuzz-on cannot be used together.[/error]'
        )
        raise typer.Exit(1)

    if generator_args and not finder:
        console.console.print(
            '[error]Option --generator/-g requires --finder/-f.[/error]'
        )
        raise typer.Exit(1)

    if finder and not generator_args and not (fuzz or fuzz_on):
        console.console.print(
            '[error]Option --finder/-f requires either --generator/-g or --fuzz/--fuzz-on.[/error]'
        )
        raise typer.Exit(1)

    fuzz_arg: Optional[Union[List[str], bool]] = None
    if fuzz_on:
        fuzz_arg = fuzz_on
    elif fuzz:
        fuzz_arg = True

    from rbx.box import stresses, tasks

    limits = tasks.get_limits_for_language(
        lang=None,
        timelimit_override=custom_timelimit,
        verification=VerificationLevel.FULL
        if double_timelimit
        else VerificationLevel.NONE,
    )

    with (
        utils.StatusProgress('Running stress...') as s,
        grading_context.stress(True),
        grading_context.cache_level(grading_context.CacheLevel.CACHE_COMPILATION),
    ):
        report = await stresses.run_stress(
            timeout,
            name=name,
            generator_call=generator_args,
            finder=finder,
            findings_limit=findings,
            progress=s,
            verbose=verbose,
            sanitized=sanitized,
            print_descriptors=print_descriptors,
            skip_invalid_testcases=skip_invalid_testcases,
            limits=limits,
            find_slowest=find_slowest,
            fuzz=fuzz_arg,
            validate=validate,
        )

    stresses.print_stress_report(report)

    if not report.findings:
        return

    # Add found tests.
    res = rich.prompt.Confirm.ask(
        'Do you want to add the tests that were found to a test group?',
        console=console.console,
    )
    if not res:
        return
    testgroup = None
    while testgroup is None or testgroup:
        groups_by_name = {
            name: group
            for name, group in package.get_test_groups_by_name().items()
            if group.generatorScript is not None
            and group.generatorScript.path.suffix == '.txt'
        }

        import questionary

        testgroup = await questionary.select(
            'Choose the testgroup to add the tests to.\nOnly test groups that have a .txt generatorScript are shown below: ',
            choices=list(groups_by_name) + ['(create new script)', '(skip)'],
        ).ask_async()

        if testgroup == '(create new script)':
            new_script_name = await questionary.text(
                'Enter the name of the new .txt generatorScript file: '
            ).ask_async()
            new_script_path = pathlib.Path(new_script_name).with_suffix('.txt')
            new_script_path.parent.mkdir(parents=True, exist_ok=True)
            new_script_path.touch()

            # Temporarily create a new testgroup with the new script.
            testgroup = new_script_path.stem
            groups_by_name[testgroup] = TestcaseGroup(
                name=testgroup, generatorScript=GeneratorScript(path=new_script_path)
            )
            ru, problem_yml = package.get_ruyaml()
            if 'testcases' not in problem_yml:
                problem_yml['testcases'] = []
            problem_yml['testcases'].append(
                {
                    'name': testgroup,
                    'generatorScript': {'path': new_script_path.name},
                }
            )
            dest = package.find_problem_yaml()
            assert dest is not None
            utils.save_ruyaml(dest, ru, problem_yml)
            package.clear_package_cache()

        if testgroup not in groups_by_name:
            break
        try:
            subgroup = groups_by_name[testgroup]
            assert subgroup.generatorScript is not None
            generator_script = pathlib.Path(subgroup.generatorScript.path)
            handler = generator_script_handlers.get_generator_script_handler(
                generator_script.read_text(),
                generator_script_handlers.GeneratorScriptHandlerParams(
                    subgroup.generatorScript,
                ),
            )

            stress_text = f'Obtained by running `rbx {shlex.join(sys.argv[1:])}`'
            handler.append(
                [finding.generator for finding in report.findings],
                comment=stress_text,
            )
            generator_script.write_text(handler.script)

            console.console.print(
                f"Added [item]{len(report.findings)}[/item] tests to test group [item]{testgroup}[/item]'s generatorScript at {subgroup.generatorScript.href()}"
            )
        except typer.Exit:
            continue
        break


@app.command(
    'compile',
    rich_help_panel='Testing',
    help='Compile an asset given its path.',
)
@package.within_problem
@syncer.sync
async def compile_command(
    path: Annotated[
        Optional[str],
        PackagePath,
        typer.Argument(help='Path to the asset to compile.'),
    ] = None,
    sanitized: bool = typer.Option(
        False,
        '--sanitized',
        '-s',
        help='Whether to compile the asset with sanitizers enabled.',
    ),
    warnings: bool = typer.Option(
        False,
        '--warnings',
        '-w',
        help='Whether to compile the asset with warnings enabled.',
    ),
    all: bool = typer.Option(
        False,
        '--all',
        '-a',
        help='Whether to compile all assets.',
    ),
):
    if path is None and not all:
        import questionary

        path = await questionary.path("What's the path to your asset?").ask_async()
        if path is None:
            console.console.print('[error]No path specified.[/error]')
            raise typer.Exit(1)

    if all:
        for solution in package.get_solutions():
            compile.any(str(solution.path), sanitized, warnings)
        if package.get_checker() is not None:
            compile.any(str(package.get_checker().path), sanitized, warnings)
        if package.get_validator() is not None:
            compile.any(str(package.get_validator().path), sanitized, warnings)
        if package.get_interactor() is not None:
            compile.any(str(package.get_interactor().path), sanitized, warnings)

    if path is not None:
        compile.any(path, sanitized, warnings)


@app.command(
    'validate',
    rich_help_panel='Testing',
    help='Run the validator in a one-off fashion, interactively.',
)
@package.within_problem
@syncer.sync
async def validate(
    path: Annotated[
        Optional[str],
        PackagePath,
        typer.Option('--path', '-p', help='Path to the testcase to validate.'),
    ] = None,
):
    all_validators = package.get_all_validators()
    if not all_validators:
        console.console.print('[error]No validator found for this problem.[/error]')
        raise typer.Exit(1)

    with utils.StatusProgress('Compiling validators...') as s:
        validators_digests = validators.compile_validators(all_validators, progress=s)

    input = console.multiline_prompt('Testcase input')

    if path is None:
        with tempfile.TemporaryDirectory() as tmpdir:
            tmppath = pathlib.Path(tmpdir) / '000.in'
            tmppath.write_text(input)

            infos = await validators.validate_one_off(
                pathlib.Path(tmppath), all_validators, validators_digests
            )
    else:
        infos = await validators.validate_one_off(
            pathlib.Path(path), all_validators, validators_digests
        )

    validators.print_validation_report(infos)


@app.command(
    'unit',
    rich_help_panel='Testing',
    help='Run unit tests for the validator and checker.',
)
@package.within_problem
@syncer.sync
async def unit_tests():
    from rbx.box import unit

    with utils.StatusProgress('Running unit tests...') as s:
        await unit.run_unit_tests(s)


@app.command(
    'header',
    rich_help_panel='Configuration',
    help='Generate the rbx.h header file.',
)
@package.within_problem
def header():
    generate_header()


# TODO: warn when using a preset (or show it)
@app.command(
    'environment, env',
    rich_help_panel='Configuration',
    help='Set or show the current box environment.',
)
def environment_command(
    env: Annotated[Optional[str], typer.Argument()] = None,
    install_from: Annotated[
        Optional[str],
        typer.Option(
            '--install',
            '-i',
            help='Whether to install this environment from the given file.',
        ),
    ] = None,
):
    if env is None:
        console.console.print(
            f'Current environment: [item]{environment.get_active_environment_description()}[/item]'
        )
        console.console.print(f'Location: {environment.get_active_environment_path()}')
        return
    if install_from is not None:
        environment.install_environment(env, pathlib.Path(install_from))
    if not get_app_environment_path(env).is_file():
        console.console.print(
            f'[error]Environment [item]{env}[/item] does not exist.[/error]'
        )
        raise typer.Exit(1)

    cfg = config.get_config()
    if env == cfg.boxEnvironment:
        console.console.print(
            f'Environment is already set to [item]{env}[/item].',
        )
        return
    console.console.print(
        f'Changing global environment from [item]{cfg.boxEnvironment}[/item] to [item]{env}[/item]...'
    )
    cfg.boxEnvironment = env
    config.save_config(cfg)

    # Also clear cache when changing environments.
    clear()


@app.command(
    'languages',
    rich_help_panel='Configuration',
    help='List the languages available in this environment',
)
def languages():
    env = environment.get_environment()

    console.console.print(
        f'[success]There are [item]{len(env.languages)}[/item] language(s) available.'
    )

    for language in env.languages:
        console.console.print(
            f'[item]{language.name}[/item], aka [item]{language.readableName or language.name}[/item]:'
        )
        console.console.print(language)
        console.console.print()


@app.command(
    'stats',
    rich_help_panel='Management',
    help='Show stats about current and related packages.',
)
@cd.within_closest_package
def stats(
    transitive: bool = typer.Option(
        False,
        '--transitive',
        '-t',
        help='Show stats about all reachable packages.',
    ),
):
    from rbx.box import stats

    if transitive:
        stats.print_reachable_package_stats()
    else:
        stats.print_package_stats()


@app.command(
    'fix',
    rich_help_panel='Management',
    help='Format files of the current package.',
)
@cd.within_closest_wrapper
def fix(print_diff: bool = typer.Option(False, '--print-diff', '-p')):
    from rbx.box import linting

    linting.fix_package(print_diff=print_diff)


@app.command(
    'wizard',
    rich_help_panel='Management',
    help='Run the wizard.',
)
@cd.within_closest_package
def wizard():
    from rbx.box.wizard.server import run_server

    run_server()


@cd.within_closest_package
def _clear_package_cache():
    console.console.print('Cleaning cache and build directories...')

    def _clean(path: pathlib.Path):
        if not path.exists():
            return
        console.console.print(f'Cleaning [item]{path}[/item]...')
        shutil.rmtree(path, ignore_errors=True)

    _clean(pathlib.Path('build'))
    if cd.is_problem_package():
        _clean(package.get_build_path())
        _clean(package.get_problem_cache_path())

    if cd.is_contest_package():
        console.console.print(
            '[warning]If you want to clear the problem caches of all problems in the contest, '
            'run [item]rbx contest each clean[/item].[/warning]'
        )


@app.command(
    'clear, clean',
    rich_help_panel='Management',
    help='Clears cache and build directories.',
)
def clear(global_cache: bool = typer.Option(False, '--global', '-g')):
    cleared = False
    if global_cache:
        console.console.print('Cleaning global cache...')
        global_package.clear_global_cache()
        cleared = True

    closest_package = cd.find_package()
    if closest_package is not None:
        _clear_package_cache()
        cleared = True

    if not cleared:
        console.console.print('[error]No cache or build directories to clean.[/error]')



import pathlib
from typing import List, Optional

import typer
from pydantic import BaseModel

from rbx import console, utils
from rbx.box import builder, checkers, testcase_extractors, validators
from rbx.box.environment import VerificationLevel, VerificationParam
from rbx.box.generation_schema import GenerationTestcaseEntry
from rbx.box.testcase_utils import (
    Testcase,
    TestcaseInteraction,
    TestcaseInteractionParsingError,
    get_best_interaction_file,
    parse_interaction,
)
from rbx.box.validators import (
    TestcaseValidationInfo,
    compile_output_validators_for_entries,
)
from rbx.grading.steps import Outcome


class StatementSample(BaseModel):
    entry: GenerationTestcaseEntry
    inputPath: pathlib.Path
    outputPath: pathlib.Path
    answerPath: Optional[pathlib.Path] = None
    explanationPath: Optional[pathlib.Path] = None
    hasOutput: bool = True
    checkOutput: bool = False
    interaction: Optional[TestcaseInteraction] = None


def _get_statement_sample_from_entry(
    entry: GenerationTestcaseEntry, explanation_suffix: Optional[str] = None
) -> StatementSample:
    input_path: pathlib.Path = utils.get_empty_sentinel_path()
    output_path: pathlib.Path = utils.get_empty_sentinel_path()
    answer_path: Optional[pathlib.Path] = None
    explanation_path: Optional[pathlib.Path] = None
    interaction: Optional[TestcaseInteraction] = None

    # Process manually provided files.
    if entry.metadata.copied_from is not None:
        input_path = entry.metadata.copied_from.inputPath
        if (
            entry.metadata.copied_from.outputPath is not None
            and entry.metadata.copied_from.outputPath.is_file()
        ):
            output_path = entry.metadata.copied_from.outputPath
            answer_path = entry.metadata.copied_from.outputPath

    # Process generated files.
    testcase = entry.metadata.copied_to
    input_path = testcase.inputPath

    if testcase.outputPath is not None and testcase.outputPath.is_file():
        output_path = testcase.outputPath
        answer_path = testcase.outputPath

    def process_additional_files(testcase: Testcase):
        nonlocal input_path, output_path, explanation_path, interaction
        if explanation_suffix is not None:
            explanation_path = testcase.inputPath.with_suffix(explanation_suffix)
            if explanation_path.is_file():
                explanation_path = explanation_path

        pin_path = testcase.inputPath.with_suffix('.pin')
        pout_path = testcase.inputPath.with_suffix('.pout')

        if pin_path.is_file():
            input_path = pin_path
        if pout_path.is_file():
            output_path = pout_path

        interaction_path = get_best_interaction_file(input_path)
        if interaction_path is not None:
            try:
                interaction = parse_interaction(interaction_path)
            except TestcaseInteractionParsingError as e:
                console.console.print(
                    f'Error parsing interactive sample: [error]{e}[/error]'
                )
                raise typer.Exit(1) from e

    process_additional_files(testcase)

    # Process statement-specific manual files
    if entry.metadata.copied_from is not None:
        out_path = entry.metadata.copied_from.inputPath.with_suffix('.out')
        if out_path.is_file():
            output_path = out_path
        out_statement = entry.metadata.copied_from.inputPath.with_suffix(
            '.out.statement'
        )
        if out_statement.is_file():
            output_path = out_statement

        process_additional_files(entry.metadata.copied_from)

    # Make all paths absolute.
    input_path = utils.abspath(input_path)
    output_path = utils.abspath(output_path)
    answer_path = utils.abspath(answer_path) if answer_path is not None else None
    explanation_path = (
        utils.abspath(explanation_path) if explanation_path is not None else None
    )

    # Whether the custom specified output should be checked.
    should_check_output = False
    if (
        answer_path is not None
        and output_path.suffix == '.out'
        and answer_path != output_path
        and output_path.is_file()
    ):
        should_check_output = True

    return StatementSample(
        entry=entry,
        inputPath=input_path,
        outputPath=output_path,
        answerPath=answer_path,
        hasOutput=output_path is not None,
        checkOutput=should_check_output,
        interaction=interaction,
        explanationPath=explanation_path,
    )


async def get_sample_entries() -> List[GenerationTestcaseEntry]:
    return await testcase_extractors.extract_generation_testcases_from_groups(
        set(['samples'])
    )


async def get_statement_samples(
    explanation_suffix: Optional[str] = None,
) -> List[StatementSample]:
    """Get the statement samples from the testcase extractors.

    This function assumes that the samples group is already built."""
    entries = await get_sample_entries()

    return [
        _get_statement_sample_from_entry(entry, explanation_suffix) for entry in entries
    ]


async def _check_sample(checker_digest: str, sample: StatementSample) -> bool:
    answer_path = sample.answerPath or utils.get_empty_sentinel_path()

    result = await checkers.check(
        checker_digest,
        run_log=None,
        testcase=Testcase(
            inputPath=sample.inputPath,
            outputPath=answer_path,
        ),
        program_output=sample.outputPath,
        skip_run_log=True,
    )

    if result.outcome != Outcome.ACCEPTED:
        output_relpath = utils.relcwd(sample.outputPath)
        console.console.print(
            f'[error]Custom output for test [item]{sample.entry}[/item] failed checker.[/error]'
        )
        console.console.print(f'[error]Path: [item]{output_relpath}[/item][/error]')
        console.console.print(f'[error]Message:[/error] {result.message}')
        console.console.print()
        return False

    return True


async def _validate_sample_outputs(
    samples: List[StatementSample],
    progress: Optional[utils.StatusProgress] = None,
) -> bool:
    def step():
        if progress is not None:
            progress.step()

    validator_to_compiled_digest = compile_output_validators_for_entries(
        [sample.entry for sample in samples]
    )

    if not validator_to_compiled_digest:
        if progress is not None:
            progress.omit()
        return True

    validation_info: List[TestcaseValidationInfo] = []

    for sample in samples:
        entry = sample.entry
        for output_validator in entry.output_validators:
            compiled_digest = validator_to_compiled_digest[str(output_validator.path)]
            ok, message, _ = await validators.validate_file(
                sample.outputPath,
                output_validator,
                compiled_digest,
                group=entry.group_entry.group,
            )
            validation_info.append(
                TestcaseValidationInfo(
                    validator=output_validator,
                    testcase=entry.group_entry,
                    generation_metadata=entry.metadata,
                    path=sample.outputPath,
                    ok=ok,
                    hit_bounds={},
                    message=message,
                )
            )
            step()

    validators.print_validation_report(validation_info, output_validation=True)

    return all(info.ok for info in validation_info)


async def build_samples(
    verification: VerificationParam,
    validate: bool,
    check_outputs_only: bool = False,
) -> bool:
    ok = True
    if not check_outputs_only:
        ok = await builder.build(
            verification=verification,
            groups=set(['samples']),
            output=None,
            validate=validate,
        )
    if not ok:
        return False
    if not validate or verification < VerificationLevel.VALIDATE.value:
        return True

    # Validate manually specified statement-only outputs.
    samples = await get_statement_samples()
    samples_to_check = [sample for sample in samples if sample.checkOutput]

    if not samples_to_check:
        return True

    with utils.StatusProgress(
        'Validating manual statement outputs for testcases...',
        'Validated [item]{processed}[/item] manual statement outputs...',
        keep=True,
    ) as s:
        ok = await _validate_sample_outputs(samples_to_check, s)

    if ok:
        with utils.StatusProgress(
            'Checking manual statement outputs for testcases...',
            'Checked [item]{processed}[/item] manual statement outputs...',
            keep=True,
        ) as s:
            checker_digest = checkers.compile_checker()
            for sample in samples_to_check:
                if not await _check_sample(checker_digest, sample):
                    ok = False
                s.step()

    if not ok:
        console.console.print(
            '[error]Some manually provided sample outputs are not considered valid answers.[/error]'
        )
        console.console.print(
            '[error]If you think these files should not be checked, use the [item].out.statement[/item] file extension (not recommended).[/error]'
        )
        console.console.print(
            '[error]You can also use either the [item]-v0[/item] or the [item]--no-validate[/item] flag to disable sample validation temporarily.[/error]'
        )
    else:
        console.console.print(
            '[success]All manual statement outputs are considered valid.[/success]'
        )
    return ok



from typing import Optional, Set

from rbx import console, utils
from rbx.box import environment, package
from rbx.box.environment import VerificationLevel
from rbx.box.generators import (
    generate_outputs_for_testcases,
    generate_testcases,
)
from rbx.box.solutions import (
    is_fast,
    print_run_report,
    run_solutions,
)
from rbx.box.testcase_extractors import extract_generation_testcases_from_groups
from rbx.box.validators import (
    check_output_from_entries,
    has_validation_errors,
    print_validation_report,
    validate_outputs_from_entries,
    validate_testcases,
)
from rbx.box.visualizers import run_visualizers_for_entries


async def build(
    verification: environment.VerificationParam,
    groups: Optional[Set[str]] = None,
    output: Optional[bool] = True,
    validate: bool = True,
    visualize: bool = False,
    is_run: bool = False,
    is_statement: bool = False,
) -> bool:
    no_main_solution_report = False
    if output is None:
        output = package.get_main_solution() is not None
        no_main_solution_report = not output

    with utils.StatusProgress(
        'Building testcases...',
        'Built [item]{processed}[/item] testcases...',
        keep=True,
    ) as s:
        await generate_testcases(s, groups=groups)

    if verification > 0 and validate:
        with utils.StatusProgress(
            'Validating testcases...',
            'Validated [item]{processed}[/item] testcases...',
            keep=True,
        ) as s:
            infos = await validate_testcases(
                s,
                groups=groups,
            )
            print_validation_report(infos)

        if has_validation_errors(infos):
            console.console.print(
                '[error]Validation failed, check the report above.[/error]'
            )
            if is_run:
                console.console.print(
                    '[error]You can use the [item]--no-validate[/item] flag to skip validation.[/error]'
                )
            else:
                console.console.print(
                    '[error]You can use the [item]-v0[/item] flag to skip validation.[/error]'
                )
            return False

    entries = await extract_generation_testcases_from_groups(groups)
    with utils.StatusProgress(
        'Building outputs for testcases...',
        'Built [item]{processed}[/item] outputs...',
        keep=True,
    ) as s:
        if output:
            await generate_outputs_for_testcases(
                [entry.group_entry for entry in entries], s
            )

    if verification > 0 and validate:
        with utils.StatusProgress(
            'Validating outputs for testcases...',
            'Validated [item]{processed}[/item] outputs...',
            keep=True,
        ) as s:
            if output:
                validation_info = await validate_outputs_from_entries(entries, s)
                print_validation_report(validation_info, output_validation=True)

    if verification > 0 and validate:
        with utils.StatusProgress(
            'Checking manual answers for testcases...',
            'Checked [item]{processed}[/item] manual answers...',
            keep=True,
        ) as s:
            if output:
                validation_info = await check_output_from_entries(entries, s)
                print_validation_report(validation_info, output_validation=True)

    if visualize:
        with utils.StatusProgress(
            'Building visualizations for testcases...',
            'Built [item]{processed}[/item] visualizations...',
            keep=True,
        ) as s:
            await run_visualizers_for_entries(entries, s)

    console.console.print(
        '[success]Problem built.[/success] '
        '[warning]Check the output for verification errors![/warning]'
    )

    if no_main_solution_report:
        console.console.print(
            '[warning]No main solution found, skipping generating samples for the statement.[/warning]'
        )

    return True


async def verify(
    verification: environment.VerificationParam, groups: Optional[Set[str]] = None
) -> bool:
    if not await build(verification=verification, groups=groups):
        return False

    if verification < VerificationLevel.FAST_SOLUTIONS.value:
        return True

    tracked_solutions = None
    if verification < VerificationLevel.ALL_SOLUTIONS.value:
        tracked_solutions = {
            str(solution.path)
            for solution in package.get_solutions()
            if is_fast(solution)
        }

    with utils.StatusProgress('Running solutions...') as s:
        solution_result = run_solutions(
            progress=s,
            tracked_solutions=tracked_solutions,
            verification=VerificationLevel(verification),
        )

    console.console.print()
    console.console.rule('[status]Run report[/status]', style='status')
    return await print_run_report(
        solution_result,
        console.console,
        VerificationLevel(verification),
    )



import asyncio

from rbx.box import code, tasks
from rbx.box.environment import VerificationLevel
from rbx.box.schema import CodeItem, Testcase
from rbx.box.testing import testing_package
from rbx.grading.judge.sandbox import SandboxBase


class TestRunSolutionOnTestcase:
    """Test suite for run_solution_on_testcase function."""

    def test_run_solution_without_explicit_language(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test that solutions without explicit language field work correctly.

        Regression test for the fix where solution.language was accessed directly
        instead of using find_language_name(), which can infer the language from
        the file extension when language is None.
        """
        # Create a simple Python program without specifying language
        py_file = testing_pkg.add_file(
            'solution.py', src='program_test/simple_hello.py'
        )
        # Create CodeItem WITHOUT setting language field - it should infer from extension
        solution = CodeItem(path=py_file)  # language not set, will be None

        # Verify language is actually None (this is the regression scenario)
        assert solution.language is None

        # Compile the solution
        compiled_digest = code.compile_item(solution)

        # Create input and output files for testcase
        input_file = testing_pkg.add_file('test.in')
        input_file.write_text('')
        output_file = testing_pkg.path('test.ans')
        output_file.write_text('Hello, World!\n')

        testcase = Testcase(
            inputPath=input_file,
            outputPath=output_file,
        )

        # Create output directory
        output_dir = testing_pkg.path('outputs')
        output_dir.mkdir(exist_ok=True)

        # Run solution on testcase - this should work without crashing
        # The fix ensures find_language_name() is used instead of solution.language
        evaluation = asyncio.run(
            tasks.run_solution_on_testcase(
                solution=solution,
                compiled_digest=compiled_digest,
                checker_digest=None,
                testcase=testcase,
                output_dir=output_dir,
                verification=VerificationLevel.NONE,
                use_retries=False,
            )
        )

        # Verify execution succeeded
        assert evaluation is not None
        assert evaluation.log is not None
        assert evaluation.log.exitcode == 0
        assert evaluation.log.exitstatus == SandboxBase.EXIT_OK

        # Verify output was generated
        output_path = output_dir / 'test.out'
        assert output_path.exists()
        assert output_path.read_text().strip() == 'Hello, World!'

    def test_run_solution_without_explicit_language_cpp(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test that C++ solutions without explicit language field work correctly."""
        # Create a simple C++ program without specifying language
        cpp_file = testing_pkg.add_file('solution.cpp', src='compile_test/simple.cpp')
        # Create CodeItem WITHOUT setting language field
        solution = CodeItem(path=cpp_file)  # language not set

        # Verify language is None
        assert solution.language is None

        # Compile the solution
        compiled_digest = code.compile_item(solution)

        # Create input and output files for testcase
        input_file = testing_pkg.add_file('test.in')
        input_file.write_text('')
        output_file = testing_pkg.path('test.ans')
        output_file.write_text('')  # Simple.cpp output varies

        testcase = Testcase(
            inputPath=input_file,
            outputPath=output_file,
        )

        # Create output directory
        output_dir = testing_pkg.path('outputs')
        output_dir.mkdir(exist_ok=True)

        # Run solution on testcase
        evaluation = asyncio.run(
            tasks.run_solution_on_testcase(
                solution=solution,
                compiled_digest=compiled_digest,
                checker_digest=None,
                testcase=testcase,
                output_dir=output_dir,
                verification=VerificationLevel.NONE,
                use_retries=False,
            )
        )

        # Verify execution succeeded
        assert evaluation is not None
        assert evaluation.log is not None
        assert evaluation.log.exitcode == 0
        assert evaluation.log.exitstatus == SandboxBase.EXIT_OK

        # Verify output was generated
        output_path = output_dir / 'test.out'
        assert output_path.exists()

    def test_run_solution_with_explicit_language_still_works(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test that solutions with explicit language field still work correctly.

        This ensures the fix doesn't break existing behavior.
        """
        # Create a Python program WITH explicit language
        py_file = testing_pkg.add_file(
            'solution.py', src='program_test/simple_hello.py'
        )
        solution = CodeItem(path=py_file, language='py')  # Explicit language

        # Verify language is set
        assert solution.language == 'py'

        # Compile the solution
        compiled_digest = code.compile_item(solution)

        # Create input and output files for testcase
        input_file = testing_pkg.add_file('test.in')
        input_file.write_text('')
        output_file = testing_pkg.path('test.ans')
        output_file.write_text('Hello, World!\n')

        testcase = Testcase(
            inputPath=input_file,
            outputPath=output_file,
        )

        # Create output directory
        output_dir = testing_pkg.path('outputs')
        output_dir.mkdir(exist_ok=True)

        # Run solution on testcase
        evaluation = asyncio.run(
            tasks.run_solution_on_testcase(
                solution=solution,
                compiled_digest=compiled_digest,
                checker_digest=None,
                testcase=testcase,
                output_dir=output_dir,
                verification=VerificationLevel.NONE,
                use_retries=False,
            )
        )

        # Verify execution succeeded
        assert evaluation is not None
        assert evaluation.log is not None
        assert evaluation.log.exitcode == 0
        assert evaluation.log.exitstatus == SandboxBase.EXIT_OK


class TestRunCommunicationSolutionOnTestcase:
    """Test suite for _run_communication_solution_on_testcase function."""

    def test_communication_solution_without_explicit_language(
        self, testing_pkg: testing_package.TestingPackage
    ):
        """Test that communication solutions without explicit language work correctly.

        Regression test for the fix in _run_communication_solution_on_testcase
        where solution.language was accessed directly instead of using
        find_language_name().
        """
        # Create a simple Python program without specifying language
        py_file = testing_pkg.add_file(
            'solution.py', src='program_test/simple_hello.py'
        )
        solution = CodeItem(path=py_file)  # language not set

        # Verify language is None
        assert solution.language is None

        # Set task type to COMMUNICATION to allow interactor
        from rbx.box.schema import TaskType

        testing_pkg.set_type(TaskType.COMMUNICATION)

        # Create a simple interactor
        interactor_content = """
import sys

# Read input
input_data = open('interactor.in').read()

# Write output
with open('interactor.out', 'w') as f:
    f.write('0\\n')  # Score

sys.exit(0)
"""
        interactor_file = testing_pkg.add_file('interactor.py')
        interactor_file.write_text(interactor_content)

        # Set interactor in package
        testing_pkg.set_interactor('interactor.py', language='py')

        # Compile both
        compiled_digest = code.compile_item(solution)
        interactor_digest = code.compile_item(testing_pkg.yml.interactor)

        # Create testcase
        input_file = testing_pkg.add_file('test.in')
        input_file.write_text('')
        output_file = testing_pkg.path('test.ans')
        output_file.write_text('')

        testcase = Testcase(
            inputPath=input_file,
            outputPath=output_file,
        )

        # Create output directory
        output_dir = testing_pkg.path('outputs')
        output_dir.mkdir(exist_ok=True)

        # Run communication solution - this tests the fix in
        # _run_communication_solution_on_testcase
        evaluation = asyncio.run(
            tasks.run_solution_on_testcase(
                solution=solution,
                compiled_digest=compiled_digest,
                checker_digest=None,
                testcase=testcase,
                output_dir=output_dir,
                interactor_digest=interactor_digest,
                verification=VerificationLevel.NONE,
                use_retries=False,
                capture_pipes=False,
            )
        )

        # Verify execution succeeded
        assert evaluation is not None
        assert evaluation.log is not None
        # Communication tasks may have different exit statuses
        assert evaluation.log.exitstatus in [
            SandboxBase.EXIT_OK,
            SandboxBase.EXIT_NONZERO_RETURN,
        ]



from textual.app import ComposeResult
from textual.screen import Screen
from textual.widgets import Footer, Header, Label


class ErrorScreen(Screen):
    BINDINGS = [('q', 'app.pop_screen', 'Quit')]

    def __init__(self, message: str):
        super().__init__()
        self.message = message

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        yield Label(self.message)

    def on_mount(self):
        self.query_one(Label).border_title = 'Error'



from TexSoup import TexSoup

from rbx.box.statements.texsoup_utils import (
    add_labels_to_tikz_nodes,
    get_tikz_node_label,
    get_top_level_labeled_tikz_nodes,
    get_top_level_tikz_nodes,
    inject_externalization_for_tikz,
    inject_in_preamble,
)


def test_get_tikz_node_label_immediately_preceding():
    latex = r"""\tikzsetnextfilename{my-label}
\begin{tikzpicture}
\end{tikzpicture}"""
    soup = TexSoup(latex)
    tikz = soup.find('tikzpicture')
    assert get_tikz_node_label(tikz)[1] == 'my-label'


def test_get_tikz_node_label_with_newlines():
    latex = r"""\tikzsetnextfilename{spaced-label}


\begin{tikzpicture}
\end{tikzpicture}"""
    soup = TexSoup(latex)
    tikz = soup.find('tikzpicture')
    assert get_tikz_node_label(tikz)[1] == 'spaced-label'


def test_get_tikz_node_label_none():
    latex = r"""\begin{tikzpicture}
\end{tikzpicture}"""
    soup = TexSoup(latex)
    tikz = soup.find('tikzpicture')
    assert get_tikz_node_label(tikz) is None


def test_get_tikz_node_label_interrupted_by_text():
    # If there's text in between, it should probably return None
    latex = r"""\tikzsetnextfilename{interrupted}
some text
\begin{tikzpicture}
\end{tikzpicture}"""
    soup = TexSoup(latex)
    tikz = soup.find('tikzpicture')
    assert get_tikz_node_label(tikz) is None


def test_get_tikz_node_label_interrupted_by_command():
    latex = r"""\tikzsetnextfilename{interrupted}
\textbf{bold}
\begin{tikzpicture}
\end{tikzpicture}"""
    soup = TexSoup(latex)
    tikz = soup.find('tikzpicture')
    assert get_tikz_node_label(tikz) is None


def test_get_tikz_node_label_with_comment():
    latex = r"""\tikzsetnextfilename{commented}
% This is a comment
\begin{tikzpicture}
\end{tikzpicture}"""
    soup = TexSoup(latex)
    tikz = soup.find('tikzpicture')
    assert get_tikz_node_label(tikz)[1] == 'commented'


def test_get_tikz_node_label_multiple():
    latex = r"""\tikzsetnextfilename{first}
\begin{tikzpicture}
\end{tikzpicture}

\tikzsetnextfilename{second}
\begin{tikzpicture}
\end{tikzpicture}"""
    soup = TexSoup(latex)
    tikzs = list(soup.find_all('tikzpicture'))
    assert len(tikzs) == 2
    assert get_tikz_node_label(tikzs[0])[1] == 'first'
    assert get_tikz_node_label(tikzs[1])[1] == 'second'


def test_inject_in_preamble_with_documentclass():
    latex = r"""\documentclass{article}
\begin{document}
Hello
\end{document}"""
    soup = TexSoup(latex)
    inject_in_preamble(soup, r'\usepackage{foo}')

    # Check that it's inserted after documentclass
    s = str(soup)
    assert r'\documentclass{article}' in s
    assert r'\usepackage{foo}' in s
    # Ensure order
    assert s.index(r'\documentclass{article}') < s.index(r'\usepackage{foo}')


def test_inject_in_preamble_no_documentclass():
    latex = r"""\begin{document}
Hello
\end{document}"""
    soup = TexSoup(latex)
    inject_in_preamble(soup, r'\usepackage{bar}')

    s = str(soup)
    assert r'\usepackage{bar}' in s
    # Should be at the start roughly
    assert s.strip().startswith(r'\usepackage{bar}')


def test_inject_externalization_for_tikz():
    latex = r"""\documentclass{article}
\begin{document}
\end{document}"""
    soup = TexSoup(latex)
    inject_externalization_for_tikz(soup)

    s = str(soup)
    assert r'\usepackage{tikz}' in s
    assert r'\usetikzlibrary{external}' in s
    assert r'\tikzexternalize[prefix=artifacts/tikz_figures/]' in s


def test_get_top_level_tikz_nodes_flat():
    latex = r"""\begin{tikzpicture}
A
\end{tikzpicture}
\begin{tikzpicture}
B
\end{tikzpicture}"""
    soup = TexSoup(latex)
    nodes = get_top_level_tikz_nodes(soup)
    assert len(nodes) == 2


def test_get_top_level_tikz_nodes_nested():
    latex = r"""\begin{tikzpicture}
    Outer
    \begin{tikzpicture}
        Inner
    \end{tikzpicture}
\end{tikzpicture}"""
    soup = TexSoup(latex)
    nodes = get_top_level_tikz_nodes(soup)
    assert len(nodes) == 1
    # Check content to be sure it's the outer one
    assert 'Outer' in str(nodes[0])


def test_get_top_level_labeled_tikz_nodes():
    latex = r"""\tikzsetnextfilename{fig1}
\begin{tikzpicture}
A
\end{tikzpicture}

\begin{tikzpicture}
B
\end{tikzpicture}

\tikzsetnextfilename{fig2}
\begin{tikzpicture}
C
\end{tikzpicture}"""
    soup = TexSoup(latex)
    labeled = get_top_level_labeled_tikz_nodes(soup)
    assert len(labeled) == 2
    assert labeled[0][2] == 'fig1'
    assert 'A' in str(labeled[0][0])
    assert labeled[1][2] == 'fig2'
    assert 'C' in str(labeled[1][0])


def test_add_labels_to_tikz_nodes():
    latex = r"""\begin{tikzpicture}
A
\end{tikzpicture}
\begin{tikzpicture}
B
\end{tikzpicture}"""
    soup = TexSoup(latex)
    add_labels_to_tikz_nodes(soup, prefix='testfig')

    s = str(soup)
    assert r'\tikzsetnextfilename{testfig_0}' in s
    assert r'\tikzsetnextfilename{testfig_1}' in s

    # Verify we can find them now
    # Re-parsing is needed here as well if we were to rely on parent pointers heavily,
    # but strictly searching might work if TexSoup updated contents internally.
    # To be safe and consistent with integration tests:
    soup = TexSoup(str(soup))
    labeled = get_top_level_labeled_tikz_nodes(soup)
    assert len(labeled) == 2
    assert labeled[0][2] == 'testfig_0'
    assert labeled[1][2] == 'testfig_1'


def test_add_labels_to_tikz_nodes_mixed():
    latex = r"""\tikzsetnextfilename{existing}
\begin{tikzpicture}
A
\end{tikzpicture}"""
    soup = TexSoup(latex)
    add_labels_to_tikz_nodes(soup, prefix='new')

    s = str(soup)
    assert r'\tikzsetnextfilename{existing}' in s

    # Existing ones are skipped, so no new labels should be added for node 0
    # But wait, the test input has only ONE node, which is already labeled.
    # So add_labels should do nothing.
    assert r'\tikzsetnextfilename{new_0}' not in s

    tikzs = list(soup.find_all('tikzpicture'))
    assert len(tikzs) == 1
    assert get_tikz_node_label(tikzs[0])[1] == 'existing'


def test_add_labels_and_replace():
    from rbx.box.statements.texsoup_utils import replace_labeled_tikz_nodes

    latex = r"""\begin{tikzpicture}
A
\end{tikzpicture}
Text
\begin{tikzpicture}
B
\end{tikzpicture}"""
    soup = TexSoup(latex)

    # 1. Add labels
    add_labels_to_tikz_nodes(soup, prefix='fig')

    # Workaround: Re-parse to fix tree integrity
    soup = TexSoup(str(soup))

    # 2. Replace labeled nodes
    replace_labeled_tikz_nodes(soup, prefix='img/', center=True)

    s = str(soup)
    assert r'\tikzsetnextfilename' not in s
    assert r'\begin{tikzpicture}' not in s
    assert r'\end{tikzpicture}' not in s

    # Check for replacements
    assert r'\begin{center}\includegraphics{img/fig_0}\end{center}' in s
    assert r'\begin{center}\includegraphics{img/fig_1}\end{center}' in s


def test_add_labels_and_replace_mixed():
    from rbx.box.statements.texsoup_utils import replace_labeled_tikz_nodes

    latex = r"""\tikzsetnextfilename{manual}
\begin{tikzpicture}
Manual
\end{tikzpicture}
\begin{tikzpicture}
Auto
\end{tikzpicture}"""
    soup = TexSoup(latex)

    # 1. Add labels (should skip 'manual' and label 'Auto')
    add_labels_to_tikz_nodes(soup, prefix='auto')

    # Workaround: Re-parse
    soup = TexSoup(str(soup))

    # 2. Replace
    replace_labeled_tikz_nodes(soup, prefix='out/', center=True)

    s = str(soup)
    assert r'\tikzsetnextfilename' not in s

    # Manual should be preserved but replaced with its label
    assert r'\begin{center}\includegraphics{out/manual}\end{center}' in s

    # Auto should get new label
    # Note: index is 1 because it's the second node in iteration
    assert r'\begin{center}\includegraphics{out/auto_1}\end{center}' in s


def test_add_labels_and_replace_no_center():
    from rbx.box.statements.texsoup_utils import replace_labeled_tikz_nodes

    latex = r"""\begin{tikzpicture}
A
\end{tikzpicture}"""
    soup = TexSoup(latex)

    add_labels_to_tikz_nodes(soup, prefix='nc')
    soup = TexSoup(str(soup))
    replace_labeled_tikz_nodes(soup, prefix='p/', center=False)

    s = str(soup)
    assert r'\begin{center}' not in s
    assert r'\includegraphics{p/nc_0}' in s



"""Tests for contest_utils (match_problem, get_problems_of_interest) with short_name and aliases."""

from typing import Optional
from unittest.mock import patch

from rbx.box.contest.contest_utils import get_problems_of_interest, match_problem
from rbx.box.contest.schema import Contest, ContestProblem


def _p(short_name: str, aliases: Optional[list[str]] = None) -> ContestProblem:
    return ContestProblem(short_name=short_name, aliases=aliases or [])


class TestMatchProblem:
    def test_wildcard_matches_all(self):
        assert match_problem('*', _p('A')) is True
        assert match_problem('*', _p('B', ['choco'])) is True

    def test_range_matches_by_short_name_only(self):
        assert match_problem('A-C', _p('B')) is True
        assert match_problem('A-C', _p('A')) is True
        assert match_problem('A-C', _p('C')) is True
        assert match_problem('A-C', _p('D')) is False
        # Range does not match by alias
        assert match_problem('A-C', _p('D', ['choco'])) is False

    def test_comma_list_matches_short_name(self):
        assert match_problem('A,B,C', _p('B')) is True
        assert match_problem('A, B , C', _p('B')) is True
        assert match_problem('A,B', _p('C')) is False

    def test_comma_list_matches_alias(self):
        assert match_problem('choco,other', _p('A', ['choco'])) is True
        assert match_problem('choco', _p('A', ['choco'])) is True
        assert match_problem('A,choco', _p('A', ['choco'])) is True
        assert match_problem('other', _p('A', ['choco'])) is False

    def test_comma_list_case_insensitive(self):
        assert match_problem('CHOCO', _p('A', ['choco'])) is True
        assert match_problem('Choco', _p('A', ['choco'])) is True
        assert match_problem('a', _p('A')) is True


class TestGetProblemsOfInterest:
    @patch('rbx.box.contest.contest_utils.contest_package.find_contest_package_or_die')
    def test_returns_problems_matching_short_name_or_alias(self, mock_find):
        contest = Contest(
            name='Test',
            problems=[
                ContestProblem(short_name='A', aliases=['apple']),
                ContestProblem(short_name='B', aliases=[]),
                ContestProblem(short_name='C', aliases=['choco', 'cake']),
            ],
        )
        mock_find.return_value = contest

        assert len(get_problems_of_interest('A')) == 1
        assert get_problems_of_interest('A')[0].short_name == 'A'

        assert len(get_problems_of_interest('apple')) == 1
        assert get_problems_of_interest('apple')[0].short_name == 'A'

        assert len(get_problems_of_interest('choco')) == 1
        assert get_problems_of_interest('choco')[0].short_name == 'C'

        two = get_problems_of_interest('A,B')
        assert len(two) == 2
        assert {p.short_name for p in two} == {'A', 'B'}

        two_by_alias = get_problems_of_interest('apple,C')
        assert len(two_by_alias) == 2
        assert {p.short_name for p in two_by_alias} == {'A', 'C'}

        all_three = get_problems_of_interest('*')
        assert len(all_three) == 3



import pathlib

from rbx.box.stressing.generator_script_parser import (
    ScriptGeneratedInput,
    parse_and_transform,
)


class TestParseAndTransformFunction:
    """Test suite for the parse_and_transform function behavior."""

    def test_parse_and_transform_returns_list_of_script_generated_inputs(self):
        """Test that parse_and_transform returns a list of ScriptGeneratedInput objects."""
        script = 'gens/generator --arg=value'
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert isinstance(result, list)
        assert len(result) == 1
        assert isinstance(result[0], ScriptGeneratedInput)

    def test_parse_and_transform_empty_script(self):
        """Test transforming empty script returns empty list."""
        script = ''
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert result == []

    def test_parse_and_transform_only_comments(self):
        """Test that comments are filtered out and don't produce results."""
        script = """
// Comment 1
// Comment 2
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert result == []

    def test_parse_and_transform_only_newlines(self):
        """Test that newlines alone don't produce results."""
        script = '\n\n\n'
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert result == []

    def test_parse_and_transform_simple_generator_call(self):
        """Test transforming a simple generator call."""
        script = 'gens/generator'
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        assert result[0].generator_call is not None
        assert result[0].generator_call.name == 'gens/generator'
        assert result[0].generator_call.args is None
        assert result[0].copied_from is None
        assert result[0].group is None

    def test_parse_and_transform_generator_call_with_args(self):
        """Test transforming generator call with arguments."""
        script = 'gens/generator --MAX_N=100 --MIN_N=30 extra'
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        gen_call = result[0].generator_call
        assert gen_call is not None
        assert gen_call.name == 'gens/generator'
        assert gen_call.args == '--MAX_N=100 --MIN_N=30 extra'

    def test_parse_and_transform_generator_call_strips_args_whitespace(self):
        """Test that argument whitespace is stripped."""
        script = 'gens/generator    --arg=value   '
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        gen_call = result[0].generator_call
        assert gen_call is not None
        assert gen_call.args == '--arg=value'

    def test_parse_and_transform_multiple_generator_calls(self):
        """Test transforming multiple generator calls."""
        script = """
gens/gen1 --A=1
gens/gen2 --B=2
gens/gen3 --C=3
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 3
        gen_call_0 = result[0].generator_call
        assert gen_call_0 is not None
        assert gen_call_0.name == 'gens/gen1'
        assert gen_call_0.args == '--A=1'
        gen_call_1 = result[1].generator_call
        assert gen_call_1 is not None
        assert gen_call_1.name == 'gens/gen2'
        assert gen_call_1.args == '--B=2'
        gen_call_2 = result[2].generator_call
        assert gen_call_2 is not None
        assert gen_call_2.name == 'gens/gen3'
        assert gen_call_2.args == '--C=3'

    def test_parse_and_transform_copy_directive(self):
        """Test transforming @copy directive."""
        script = '@copy tests/manual.in'
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        assert result[0].copied_from is not None
        assert result[0].copied_from.inputPath == pathlib.Path('tests/manual.in')
        assert result[0].generator_call is None
        assert result[0].group is None

    def test_parse_and_transform_multiple_copy_directives(self):
        """Test transforming multiple @copy directives."""
        script = """
@copy tests/sample1.in
@copy tests/sample2.in
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 2
        copied_0 = result[0].copied_from
        assert copied_0 is not None
        assert copied_0.inputPath == pathlib.Path('tests/sample1.in')
        copied_1 = result[1].copied_from
        assert copied_1 is not None
        assert copied_1.inputPath == pathlib.Path('tests/sample2.in')

    def test_parse_and_transform_tracks_line_numbers(self):
        """Test that line numbers are tracked correctly in generator_script."""
        script = """
// Comment on line 2
gens/gen1 --A=1
gens/gen2 --B=2
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 2
        gen_script_0 = result[0].generator_script
        assert gen_script_0 is not None
        assert gen_script_0.path == script_path
        assert gen_script_0.line == 3
        gen_script_1 = result[1].generator_script
        assert gen_script_1 is not None
        assert gen_script_1.path == script_path
        assert gen_script_1.line == 4

    def test_parse_and_transform_tracks_copy_line_numbers(self):
        """Test that line numbers are tracked for @copy directives."""
        script = """
@copy tests/sample1.in
@copy tests/sample2.in
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        gen_script_0 = result[0].generator_script
        assert gen_script_0 is not None
        assert gen_script_0.line == 2
        gen_script_1 = result[1].generator_script
        assert gen_script_1 is not None
        assert gen_script_1.line == 3

    def test_parse_and_transform_inline_input_single_quotes(self):
        """Test transforming @input directive with single-quoted string."""
        script = "@input 'test data'"
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        assert result[0].content == 'test data'
        assert result[0].generator_call is None
        assert result[0].copied_from is None
        assert result[0].group is None

    def test_parse_and_transform_inline_input_double_quotes(self):
        """Test transforming @input directive with double-quoted string."""
        script = '@input "test data"'
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        assert result[0].content == 'test data'

    def test_parse_and_transform_inline_input_with_escape_sequences(self):
        """Test transforming @input directive with escape sequences."""
        script = r"@input '123\n456\n789\n'"
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        assert result[0].content == '123\n456\n789\n'

    def test_parse_and_transform_inline_input_with_escaped_quotes(self):
        """Test transforming @input directive with escaped quotes."""
        script = r"@input 'it\'s a test'"
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        assert result[0].content == "it's a test"

    def test_parse_and_transform_inline_input_triple_quoted(self):
        """Test transforming @input directive with triple-quoted string."""
        script = '''@input """
123
456
789
"""'''
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        # Triple-quoted strings have leading/trailing empty lines stripped
        assert result[0].content == '123\n456\n789\n'

    def test_parse_and_transform_multiple_inline_inputs(self):
        """Test transforming multiple @input directives."""
        script = """
@input 'first input'
@input "second input"
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 2
        assert result[0].content == 'first input'
        assert result[1].content == 'second input'

    def test_parse_and_transform_tracks_inline_input_line_numbers(self):
        """Test that line numbers are tracked for @input directives."""
        script = """
@input 'first'
@input "second"
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        gen_script_0 = result[0].generator_script
        assert gen_script_0 is not None
        assert gen_script_0.line == 2
        gen_script_1 = result[1].generator_script
        assert gen_script_1 is not None
        assert gen_script_1.line == 3

    def test_parse_and_transform_inline_input_in_testgroup(self):
        """Test transforming @input directive inside a testgroup."""
        script = """
@testgroup group1 {
    @input 'test data in group'
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        assert result[0].group == 'group1'
        assert result[0].content == 'test data in group'

    def test_parse_and_transform_inline_input_brace_block(self):
        """Test transforming @input directive with brace block syntax."""
        script = """@input {
1 2 3
4 5 6
}"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        assert result[0].content == '1 2 3\n4 5 6\n'

    def test_parse_and_transform_inline_input_brace_block_empty(self):
        """Test transforming @input directive with empty brace block."""
        script = """@input {
}"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        # Empty brace block produces empty string
        assert result[0].content == ''

    def test_parse_and_transform_inline_input_brace_vs_triple_quote(self):
        """Test that brace block and triple-quote produce similar results."""
        script = """@input {
1 2 3
4 5 6
}

@input \"\"\"
1 2 3
4 5 6
\"\"\"
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 2
        # Both normalize whitespace the same way now (strip leading/trailing empty lines)
        assert result[0].content == '1 2 3\n4 5 6\n'
        assert result[1].content == '1 2 3\n4 5 6\n'
        # They're now equivalent
        assert result[0].content == result[1].content

    def test_parse_and_transform_inline_input_brace_with_whitespace(self):
        """Test transforming @input brace block with various whitespace."""
        script = """@input {
  indented line
\ttab line
mixed\twhitespace
}"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        # Whitespace normalization strips leading whitespace from each line
        assert result[0].content == 'indented line\ntab line\nmixed\twhitespace\n'

    def test_parse_and_transform_inline_input_brace_in_testgroup(self):
        """Test transforming @input brace block inside a testgroup."""
        script = """
@testgroup group1 {
    @input {
abc
def
    }
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        assert result[0].group == 'group1'
        assert result[0].content == 'abc\ndef\n'

    def test_parse_and_transform_mixed_input_syntaxes(self):
        """Test mixing all three input syntaxes."""
        script = """
@input 'single quoted'
@input "double quoted"
@input \"\"\"
triple
quoted
\"\"\"
@input {
brace
block
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 4
        assert result[0].content == 'single quoted'
        assert result[1].content == 'double quoted'
        # Triple-quoted has leading/trailing empty lines stripped
        assert result[2].content == 'triple\nquoted\n'
        # Brace block has whitespace normalized
        assert result[3].content == 'brace\nblock\n'

    def test_parse_and_transform_simple_testgroup(self):
        """Test transforming a simple testgroup."""
        script = """
@testgroup group1 {
    gens/generator --X=5
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        assert result[0].group == 'group1'
        gen_call = result[0].generator_call
        assert gen_call is not None
        assert gen_call.name == 'gens/generator'
        assert gen_call.args == '--X=5'

    def test_parse_and_transform_testgroup_with_multiple_statements(self):
        """Test transforming testgroup with multiple statements."""
        script = """
@testgroup my_group {
    gens/gen1 --A=1
    gens/gen2 --B=2
    @copy tests/manual.in
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 3
        assert all(item.group == 'my_group' for item in result)
        gen_call_0 = result[0].generator_call
        assert gen_call_0 is not None
        assert gen_call_0.name == 'gens/gen1'
        gen_call_1 = result[1].generator_call
        assert gen_call_1 is not None
        assert gen_call_1.name == 'gens/gen2'
        copied_2 = result[2].copied_from
        assert copied_2 is not None
        assert copied_2.inputPath == pathlib.Path('tests/manual.in')

    def test_parse_and_transform_testgroup_filters_comments(self):
        """Test that comments inside testgroups are filtered out."""
        script = """
@testgroup group1 {
    // This is a comment
    gens/gen1 --A=1
    // Another comment
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        assert result[0].group == 'group1'

    def test_parse_and_transform_nested_testgroups(self):
        """Test transforming nested testgroups."""
        script = """
@testgroup outer {
    gens/gen1 --X=1
    @testgroup inner {
        gens/gen2 --Y=2
    }
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 2
        assert result[0].group == 'outer'
        gen_call_0 = result[0].generator_call
        assert gen_call_0 is not None
        assert gen_call_0.name == 'gens/gen1'
        assert result[1].group == 'inner'
        gen_call_1 = result[1].generator_call
        assert gen_call_1 is not None
        assert gen_call_1.name == 'gens/gen2'

    def test_parse_and_transform_nested_testgroups_inner_takes_precedence(self):
        """Test that inner testgroup name takes precedence over outer."""
        script = """
@testgroup outer {
    @testgroup inner {
        gens/gen1 --X=1
    }
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 1
        # Inner group name should take precedence
        assert result[0].group == 'inner'

    def test_parse_and_transform_multiple_top_level_testgroups(self):
        """Test transforming multiple top-level testgroups."""
        script = """
@testgroup group1 {
    gens/gen1 --A=1
}

@testgroup group2 {
    gens/gen2 --B=2
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 2
        assert result[0].group == 'group1'
        assert result[1].group == 'group2'

    def test_parse_and_transform_empty_testgroup(self):
        """Test transforming empty testgroup."""
        script = """
@testgroup empty_group {
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert result == []

    def test_parse_and_transform_mixed_grouped_and_ungrouped(self):
        """Test transforming script with both grouped and ungrouped statements."""
        script = """
gens/gen_ungrouped --X=1

@testgroup group1 {
    gens/gen_grouped --Y=2
}

@copy tests/manual.in
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 3
        assert result[0].group is None
        gen_call_0 = result[0].generator_call
        assert gen_call_0 is not None
        assert gen_call_0.name == 'gens/gen_ungrouped'
        assert result[1].group == 'group1'
        gen_call_1 = result[1].generator_call
        assert gen_call_1 is not None
        assert gen_call_1.name == 'gens/gen_grouped'
        assert result[2].group is None
        assert result[2].copied_from is not None

    def test_parse_and_transform_testgroup_with_hyphenated_name(self):
        """Test transforming testgroup with hyphenated name."""
        script = """
@testgroup test-group-name {
    gens/gen --X=1
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert result[0].group == 'test-group-name'

    def test_parse_and_transform_testgroup_with_underscore_name(self):
        """Test transforming testgroup with underscore in name."""
        script = """
@testgroup test_group_name {
    gens/gen --X=1
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert result[0].group == 'test_group_name'

    def test_parse_and_transform_testgroup_with_numbers(self):
        """Test transforming testgroup with numbers in name."""
        script = """
@testgroup group123 {
    gens/gen --X=1
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert result[0].group == 'group123'

    def test_parse_and_transform_preserves_script_path(self):
        """Test that script path is preserved in all results."""
        script = """
gens/gen1
@copy tests/manual.in
"""
        script_path = pathlib.Path('custom/path/script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 2
        gen_script_0 = result[0].generator_script
        assert gen_script_0 is not None
        assert gen_script_0.path == script_path
        gen_script_1 = result[1].generator_script
        assert gen_script_1 is not None
        assert gen_script_1.path == script_path

    def test_parse_and_transform_generator_script_entry_str(self):
        """Test that GeneratorScriptEntry str representation is correct."""
        script = 'gens/gen --X=1'
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert str(result[0].generator_script) == 'test_script.txt:1'

    def test_parse_and_transform_complex_real_world_script(self):
        """Test transforming a complex real-world style script."""
        script = """
// Generate small test cases
@testgroup small {
    gens/gen_random --MAX_N=10 --seed=1
    gens/gen_random --MAX_N=10 --seed=2
    gens/gen_edge --type=min
}

// Copy manual tests
@copy tests/manual/edge1.in
@copy tests/manual/edge2.in

// Generate large test cases
@testgroup large {
    gens/gen_random --MAX_N=100000 --seed=42
    gens/gen_edge --type=max
}

// Final edge case
gens/gen_special --config=final
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 8

        # Check small group
        small_items = [r for r in result if r.group == 'small']
        assert len(small_items) == 3
        small_gen_call_0 = small_items[0].generator_call
        assert small_gen_call_0 is not None
        assert small_gen_call_0.name == 'gens/gen_random'
        assert small_gen_call_0.args == '--MAX_N=10 --seed=1'

        # Check copied items
        copied_items = [r for r in result if r.copied_from is not None]
        assert len(copied_items) == 2
        copied_0 = copied_items[0].copied_from
        assert copied_0 is not None
        assert copied_0.inputPath == pathlib.Path('tests/manual/edge1.in')
        copied_1 = copied_items[1].copied_from
        assert copied_1 is not None
        assert copied_1.inputPath == pathlib.Path('tests/manual/edge2.in')

        # Check large group
        large_items = [r for r in result if r.group == 'large']
        assert len(large_items) == 2

        # Check ungrouped item
        ungrouped_items = [
            r for r in result if r.group is None and r.copied_from is None
        ]
        assert len(ungrouped_items) == 1
        ungrouped_gen_call = ungrouped_items[0].generator_call
        assert ungrouped_gen_call is not None
        assert ungrouped_gen_call.name == 'gens/gen_special'

    def test_parse_and_transform_maintains_order(self):
        """Test that transformation maintains statement order."""
        script = """
gens/gen1
gens/gen2
gens/gen3
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 3
        gen_call_0 = result[0].generator_call
        assert gen_call_0 is not None
        assert gen_call_0.name == 'gens/gen1'
        gen_call_1 = result[1].generator_call
        assert gen_call_1 is not None
        assert gen_call_1.name == 'gens/gen2'
        gen_call_2 = result[2].generator_call
        assert gen_call_2 is not None
        assert gen_call_2.name == 'gens/gen3'

    def test_parse_and_transform_testgroup_maintains_order(self):
        """Test that testgroup transformation maintains statement order."""
        script = """
@testgroup group1 {
    gens/gen1
    @copy tests/test1.in
    gens/gen2
    @copy tests/test2.in
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 4
        gen_call_0 = result[0].generator_call
        assert gen_call_0 is not None
        assert gen_call_0.name == 'gens/gen1'
        copied_1 = result[1].copied_from
        assert copied_1 is not None
        assert copied_1.inputPath == pathlib.Path('tests/test1.in')
        gen_call_2 = result[2].generator_call
        assert gen_call_2 is not None
        assert gen_call_2.name == 'gens/gen2'
        copied_3 = result[3].copied_from
        assert copied_3 is not None
        assert copied_3.inputPath == pathlib.Path('tests/test2.in')

    def test_parse_and_transform_line_numbers_with_comments(self):
        """Test that line numbers account for comments correctly."""
        script = """
// Line 2
// Line 3
gens/gen1
// Line 5
gens/gen2
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 2
        gen_script_0 = result[0].generator_script
        assert gen_script_0 is not None
        assert gen_script_0.line == 4
        gen_script_1 = result[1].generator_script
        assert gen_script_1 is not None
        assert gen_script_1.line == 6

    def test_parse_and_transform_deeply_nested_testgroups(self):
        """Test transforming deeply nested testgroups."""
        script = """
@testgroup level1 {
    gens/gen1
    @testgroup level2 {
        gens/gen2
        @testgroup level3 {
            gens/gen3
        }
    }
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 3
        assert result[0].group == 'level1'
        assert result[1].group == 'level2'
        assert result[2].group == 'level3'

    def test_parse_and_transform_generator_with_various_filepath_formats(self):
        """Test transforming generators with different filepath formats."""
        script = """
gen
gen.cpp
gens/gen
path/to/generator.py
../gen
../../parent/gen.py
./gen
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 7
        gen_call_0 = result[0].generator_call
        assert gen_call_0 is not None
        assert gen_call_0.name == 'gen'
        gen_call_1 = result[1].generator_call
        assert gen_call_1 is not None
        assert gen_call_1.name == 'gen.cpp'
        gen_call_2 = result[2].generator_call
        assert gen_call_2 is not None
        assert gen_call_2.name == 'gens/gen'
        gen_call_3 = result[3].generator_call
        assert gen_call_3 is not None
        assert gen_call_3.name == 'path/to/generator.py'
        gen_call_4 = result[4].generator_call
        assert gen_call_4 is not None
        assert gen_call_4.name == '../gen'
        gen_call_5 = result[5].generator_call
        assert gen_call_5 is not None
        assert gen_call_5.name == '../../parent/gen.py'
        gen_call_6 = result[6].generator_call
        assert gen_call_6 is not None
        assert gen_call_6.name == './gen'

    def test_parse_and_transform_copy_with_various_path_formats(self):
        """Test transforming @copy with different relative path formats."""
        script = """
@copy test.in
@copy tests/test.in
@copy path/to/deep/test.in
@copy ../test.in
@copy ../../parent/test.in
@copy ./test.in
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 6
        copied_0 = result[0].copied_from
        assert copied_0 is not None
        assert copied_0.inputPath == pathlib.Path('test.in')
        copied_1 = result[1].copied_from
        assert copied_1 is not None
        assert copied_1.inputPath == pathlib.Path('tests/test.in')
        copied_2 = result[2].copied_from
        assert copied_2 is not None
        assert copied_2.inputPath == pathlib.Path('path/to/deep/test.in')
        copied_3 = result[3].copied_from
        assert copied_3 is not None
        assert copied_3.inputPath == pathlib.Path('../test.in')
        copied_4 = result[4].copied_from
        assert copied_4 is not None
        assert copied_4.inputPath == pathlib.Path('../../parent/test.in')
        copied_5 = result[5].copied_from
        assert copied_5 is not None
        assert copied_5.inputPath == pathlib.Path('./test.in')

    def test_parse_and_transform_script_with_blank_lines(self):
        """Test that blank lines don't affect transformation."""
        script = """

gens/gen1


gens/gen2


"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 2

    def test_parse_and_transform_testgroup_line_numbers(self):
        """Test that line numbers inside testgroups are correct."""
        script = """
@testgroup group1 {
    gens/gen1
    gens/gen2
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 2
        gen_script_0 = result[0].generator_script
        assert gen_script_0 is not None
        assert gen_script_0.line == 3
        gen_script_1 = result[1].generator_script
        assert gen_script_1 is not None
        assert gen_script_1.line == 4

    def test_parse_and_transform_mixed_line_numbers(self):
        """Test line numbers in complex mixed script."""
        script = """
gens/gen1
@testgroup group1 {
    gens/gen2
}
@copy tests/test.in
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 3
        gen_script_0 = result[0].generator_script
        assert gen_script_0 is not None
        assert gen_script_0.line == 2
        gen_script_1 = result[1].generator_script
        assert gen_script_1 is not None
        assert gen_script_1.line == 4
        gen_script_2 = result[2].generator_script
        assert gen_script_2 is not None
        assert gen_script_2.line == 6

    def test_parse_and_transform_example_from_main_block(self):
        """Test the example script from the main block of the module."""
        script = """
// This is a comment
gens/generator --MAX_N=100 --MIN_N=30 abcdef

@copy test/in/disk.in

@testgroup my-group {
    // Comment inside group
    gens/generator2 --X=5
    gens/generator3 --Y=10 --Z=20 some more args
}

@testgroup group_2 {
    gens/generator4 --A=1
}
"""
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        assert len(result) == 5

        # First generator (ungrouped)
        gen_call_0 = result[0].generator_call
        assert gen_call_0 is not None
        assert gen_call_0.name == 'gens/generator'
        assert gen_call_0.args == '--MAX_N=100 --MIN_N=30 abcdef'
        assert result[0].group is None

        # Copy directive (ungrouped)
        copied_1 = result[1].copied_from
        assert copied_1 is not None
        assert copied_1.inputPath == pathlib.Path('test/in/disk.in')
        assert result[1].group is None

        # my-group items
        gen_call_2 = result[2].generator_call
        assert gen_call_2 is not None
        assert gen_call_2.name == 'gens/generator2'
        assert result[2].group == 'my-group'
        gen_call_3 = result[3].generator_call
        assert gen_call_3 is not None
        assert gen_call_3.name == 'gens/generator3'
        assert result[3].group == 'my-group'

        # group_2 item
        gen_call_4 = result[4].generator_call
        assert gen_call_4 is not None
        assert gen_call_4.name == 'gens/generator4'
        assert result[4].group == 'group_2'

    def test_parse_and_transform_args_not_split_as_separate_calls(self):
        """Regression test: ensure args after generator name aren't parsed as separate calls.

        Previously, a line like 'gen1 456' would be incorrectly parsed as two calls:
        - gen1 (no args)
        - 456 (as a generator name)

        This test ensures 'gen1 456' is correctly parsed as one call with name='gen1' and args='456'.
        """
        # Using explicit line construction to avoid any whitespace issues
        script = 'gen1 123\ngen1 456'
        script_path = pathlib.Path('test_script.txt')

        result = parse_and_transform(script, script_path)

        # Should be 2 generator calls, not 4+
        assert len(result) == 2

        # First call - numeric arg
        gen_call_0 = result[0].generator_call
        assert gen_call_0 is not None
        assert gen_call_0.name == 'gen1'
        assert gen_call_0.args == '123'

        # Second call - numeric arg
        gen_call_1 = result[1].generator_call
        assert gen_call_1 is not None
        assert gen_call_1.name == 'gen1'
        assert gen_call_1.args == '456'



import pathlib
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple

from rbx import utils
from rbx.box import naming
from rbx.box.contest import contest_package
from rbx.box.contest.schema import ContestStatement
from rbx.box.exception import RbxException
from rbx.box.fields import Primitive
from rbx.box.statements import statement_utils
from rbx.box.statements.builders import StatementBuilderContest
from rbx.box.statements.schema import ConversionStep, ConversionType, Statement


class StatementInheritanceError(RbxException):
    pass


@dataclass
class StatementOverrideData:
    root: pathlib.Path
    assets: List[Tuple[pathlib.Path, pathlib.Path]]
    params: Dict[ConversionType, ConversionStep]
    vars: Dict[str, Any]
    samples: bool
    inheritedFrom: Optional[ContestStatement] = None

    def to_kwargs(self, custom_vars: Dict[str, Any]) -> Dict[str, Any]:
        extra_vars = dict(self.vars)
        extra_vars.update(custom_vars or {})
        return {
            'overridden_params_root': self.root,
            'overridden_assets': self.assets,
            'overridden_params': self.params,
            'custom_vars': extra_vars,
            'inherited_from': self.inheritedFrom,
        }


def get_overrides(
    statement: ContestStatement, inherit: bool = False
) -> StatementOverrideData:
    override = statement.inheritOverride if inherit else statement.override
    contest_cwd_absolute = utils.abspath(contest_package.find_contest())
    # TODO: rethink if we should be adding statement assets into the problem context.
    contest_assets = statement_utils.get_relative_assets(
        statement.path,
        statement.assets,
        root=contest_cwd_absolute,
    )
    overridden_params = {}
    if override is not None:
        overridden_params.update({cfg.type: cfg for cfg in override.configure})
    return StatementOverrideData(
        root=contest_cwd_absolute,
        assets=contest_assets,
        params=overridden_params,
        vars=override.vars if override is not None else {},
        samples=statement.samples,
        inheritedFrom=statement if inherit else None,
    )


def get_inheritance_overrides(statement: Statement) -> StatementOverrideData:
    contest = contest_package.find_contest_package()
    if contest is None:
        with StatementInheritanceError() as e:
            e.print(
                f'[error][item]{statement.name}[/item] inherits its configuration from the contest, but no contest was found.[/error]'
            )
        raise e

    def matches(contest_statement: ContestStatement) -> bool:
        if contest_statement.joiner is None:
            return False
        if contest_statement.match is None:
            return statement.language == contest_statement.language
        return statement.name == contest_statement.match

    for contest_statement in contest.expanded_statements:
        if matches(contest_statement):
            return get_overrides(contest_statement, inherit=True)

    with StatementInheritanceError() as e:
        e.print(
            f'[error][item]{statement.name}[/item] inherits its configuration from the contest, but no matching statement was found in the contest.[/error]'
        )
    raise e


def get_statement_builder_contest_for_problem(
    language: Optional[str] = None,
    inherited_from: Optional[ContestStatement] = None,
    custom_vars: Optional[Dict[str, Primitive]] = None,
) -> Optional[StatementBuilderContest]:
    contest = contest_package.find_contest_package()
    if contest is None:
        return None
    return StatementBuilderContest(
        title=naming.get_contest_title(
            lang=inherited_from.language if inherited_from is not None else language,
            statement=inherited_from,
            contest=contest,
        ),
        problems=[],
        vars={
            **contest.expanded_vars,
            **(inherited_from.expanded_vars if inherited_from is not None else {}),
            **(custom_vars or {}),
        },
    )



# from typing import Iterable, List

# import typer
# from pydantic import BaseModel

# from rbx import console
# from rbx.box import code, package
# from rbx.box.packaging.boca.boca_language_utils import (
#     get_boca_language_from_rbx_language,
# )
# from rbx.box.packaging.boca.boca_outcome_utils import simplify_rbx_expected_outcome
# from rbx.box.schema import ExpectedOutcome, Solution
# from rbx.box.tooling.boca.scraper import BocaDetailedRun, BocaScraper, ContestSnapshot


# def _get_all_eligible_solutions() -> List[Solution]:
#     return package.get_solutions()


# class JudgedSolution(BaseModel):
#     solution: Solution
#     run: BocaDetailedRun
#     expected_outcome: ExpectedOutcome


# def submit_all_solutions(scraper: BocaScraper) -> Iterable[JudgedSolution]:
#     scraper.login()
#     if not scraper.loggedIn:
#         console.console.print('[error]Failed to login to BOCA.[/error]')
#         raise typer.Exit(1)

#     solutions = _get_all_eligible_solutions()
#     for solution in solutions:
#         boca_language = get_boca_language_from_rbx_language(
#             code.find_language_name(solution)
#         )
#         scraper.submit_as_judge(problem_index, language_index, solution.path)

#     scraper.wait_for_all_judged()
#     runs = scraper.retrieve_runs(only_judged=True)
#     runs_snapshot = ContestSnapshot(detailed_runs=runs)
#     for solution in solutions:
#         run = runs_snapshot.get_detailed_run_by_path(solution.path)
#         expected_outcome = simplify_rbx_expected_outcome(solution.outcome)
#         judged_solution = JudgedSolution(
#             solution=solution,
#             run=run,
#             expected_outcome=expected_outcome,
#         )
#         yield judged_solution



import contextlib
from typing import Optional, Tuple

from rbx import console as rconsole
from rbx.box.sanitizers.issue_stack import (
    Issue,
    IssueAccumulator,
    IssueLevel,
    add_issue,
    get_issue_accumulator,
    get_issue_stack,
    issue_level_var,
    issue_stack_var,
    pop_issue_accumulator,
    print_current_report,
    push_issue_accumulator,
)


class DummyIssue(Issue):
    def __init__(
        self,
        *,
        detailed_section: Optional[Tuple[str, ...]] = None,
        overview_section: Optional[Tuple[str, ...]] = None,
        detailed_message: str = '',
        overview_message: str = '',
    ):
        self._detailed_section = detailed_section
        self._overview_section = overview_section
        self._detailed_message = detailed_message
        self._overview_message = overview_message

    def get_detailed_section(self) -> Optional[Tuple[str, ...]]:
        return self._detailed_section

    def get_overview_section(self) -> Optional[Tuple[str, ...]]:
        return self._overview_section

    def get_detailed_message(self) -> str:
        return self._detailed_message

    def get_overview_message(self) -> str:
        return self._overview_message


@contextlib.contextmanager
def _fresh_issue_stack():
    token = issue_stack_var.set([IssueAccumulator()])
    try:
        yield
    finally:
        issue_stack_var.reset(token)


def _capture_print(fn):
    with rconsole.console.capture() as capture:
        fn()
    return capture.get()


def test_detailed_report_prints_nested_sections_and_messages():
    with _fresh_issue_stack():
        acc = get_issue_accumulator()
        acc.add_issue(
            DummyIssue(
                detailed_section=('Group1', 'Sub1'), detailed_message='Message A'
            )
        )
        acc.add_issue(
            DummyIssue(
                detailed_section=('Group1', 'Sub2'), detailed_message='Message B'
            )
        )
        acc.add_issue(
            DummyIssue(detailed_section=('Group2',), detailed_message='Message C')
        )

        output = _capture_print(acc.print_detailed_report)

        assert 'Issues' in output
        assert 'Group1' in output
        assert 'Message A' in output and 'Message B' in output and 'Message C' in output


def test_overview_report_prints_only_overview_sections():
    with _fresh_issue_stack():
        acc = get_issue_accumulator()
        acc.add_issue(
            DummyIssue(
                overview_section=('Summary', 'Outer'), overview_message='Overview 1'
            )
        )
        acc.add_issue(
            DummyIssue(
                overview_section=('Summary', 'Inner'), overview_message='Overview 2'
            )
        )
        acc.add_issue(
            DummyIssue(
                overview_section=None,
                detailed_section=('WillOnlyShowInDetailed',),
                detailed_message='SHOULD_NOT_APPEAR',
            )
        )

        # Ensure overview only shows overview issues
        output = _capture_print(acc.print_overview_report)

        assert 'Issues' in output
        assert 'Summary' in output
        assert 'Overview 1' in output and 'Overview 2' in output
        assert 'SHOULD_NOT_APPEAR' not in output


def test_add_issue_propagates_to_all_accumulators_in_stack():
    with _fresh_issue_stack():
        base_stack = list(get_issue_stack())
        assert len(base_stack) == 1
        base_acc = base_stack[0]

        push_issue_accumulator()
        top_acc = get_issue_accumulator()

        issue = DummyIssue(
            detailed_section=('Both',), detailed_message='Propagated Message'
        )
        add_issue(issue)

        out_base = _capture_print(base_acc.print_detailed_report)
        out_top = _capture_print(top_acc.print_detailed_report)

        assert 'Propagated Message' in out_base
        assert 'Propagated Message' in out_top

        # Clean up the stack change for safety within the context
        pop_issue_accumulator()


def test_print_current_report_respects_issue_level():
    with _fresh_issue_stack():
        acc = get_issue_accumulator()
        acc.add_issue(
            DummyIssue(
                detailed_section=('Sec',),
                overview_section=('Sec',),
                detailed_message='DetailedMsg',
                overview_message='OverviewMsg',
            )
        )

        # Default level is DETAILED
        out_detailed = _capture_print(print_current_report)
        assert 'DetailedMsg' in out_detailed and 'OverviewMsg' not in out_detailed

        # Switch to OVERVIEW and ensure it changes
        token = issue_level_var.set(IssueLevel.OVERVIEW)
        try:
            out_overview = _capture_print(print_current_report)
            assert 'OverviewMsg' in out_overview and 'DetailedMsg' not in out_overview
        finally:
            issue_level_var.reset(token)



import pathlib

import pytest
import typer

from rbx.box import validators as validators_mod
from rbx.box.generators import (
    ValidationError,
    generate_standalone,
    generate_testcases,
)
from rbx.box.schema import CodeItem, GeneratorCall, Testcase
from rbx.box.testcase_extractors import GenerationMetadata
from rbx.box.testing import testing_package
from rbx.grading import steps


async def test_generator_in_testplan(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_plan(
        'main',
        """
gens/gen.cpp 123
gens/gen.cpp 424242
""",
    )
    await generate_testcases()

    testing_pkg.print_debug()

    assert (
        testing_pkg.get_build_testgroup_path('main') / '000.in'
    ).read_text() == '123\n'
    assert (
        testing_pkg.get_build_testgroup_path('main') / '001.in'
    ).read_text() == '424242\n'


async def test_generator_in_testplan_without_extension(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_plan('main', 'gens/gen 123\n')
    await generate_testcases()

    testing_pkg.print_debug()

    assert (
        testing_pkg.get_build_testgroup_path('main') / '000.in'
    ).read_text() == '123\n'


async def test_aliased_generator_in_testplan(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.add_generator(
        'gens/gen.cpp', alias='gen_alias', src='generators/gen-id.cpp'
    )
    testing_pkg.add_testgroup_from_plan('main', 'gen_alias 123\n')
    await generate_testcases()

    testing_pkg.print_debug()

    assert (
        testing_pkg.get_build_testgroup_path('main') / '000.in'
    ).read_text() == '123\n'


async def test_generator_in_testplan_with_multiple_generators(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_generator('gens/gen2.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_plan(
        'main',
        """
gens/gen.cpp 123
gens/gen2.cpp 424242
""",
    )
    await generate_testcases()

    testing_pkg.print_debug()

    assert (
        testing_pkg.get_build_testgroup_path('main') / '000.in'
    ).read_text() == '123\n'
    assert (
        testing_pkg.get_build_testgroup_path('main') / '001.in'
    ).read_text() == '424242\n'


async def test_generator_with_multiple_groups(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_plan(
        'main',
        """
gens/gen.cpp 123
""",
    )
    testing_pkg.add_testgroup_from_plan(
        'secondary',
        """
gens/gen.cpp 424242
""",
    )
    await generate_testcases()

    testing_pkg.print_debug()

    assert (
        testing_pkg.get_build_testgroup_path('main') / '000.in'
    ).read_text() == '123\n'
    assert (
        testing_pkg.get_build_testgroup_path('secondary') / '000.in'
    ).read_text() == '424242\n'


async def test_generator_with_script(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_script(
        'main',
        """
for i in range(3):
    print(f'gens/gen.cpp 123 {i}')
print('gens/gen.cpp 456')
""",
    )
    await generate_testcases()

    testing_pkg.print_debug()

    assert (
        testing_pkg.get_build_testgroup_path('main') / '000.in'
    ).read_text() == '123\n'
    assert (
        testing_pkg.get_build_testgroup_path('main') / '001.in'
    ).read_text() == '123\n'
    assert (
        testing_pkg.get_build_testgroup_path('main') / '002.in'
    ).read_text() == '123\n'
    assert (
        testing_pkg.get_build_testgroup_path('main') / '003.in'
    ).read_text() == '456\n'


async def test_generator_only_necessary_groups(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_plan(
        'main',
        """
gens/gen.cpp 123
""",
    )
    testing_pkg.add_testgroup_from_plan(
        'secondary',
        """
gens/gen.cpp 424242
""",
    )
    testing_pkg.add_testgroup_from_plan(
        'non_existent',
        """
gens/gen_non_existent.cpp 424242
""",
    )
    await generate_testcases(groups={'main'})

    testing_pkg.print_debug()

    assert (
        testing_pkg.get_build_testgroup_path('main') / '000.in'
    ).read_text() == '123\n'
    assert not (testing_pkg.get_build_testgroup_path('secondary') / '000.in').exists()


async def test_generator_with_glob_and_plan(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_glob('samples', 'manual_tests/*.in')
    testing_pkg.add_testgroup_from_plan('main', 'gens/gen.cpp 424242')
    testing_pkg.add_file('manual_tests/000.in').write_text('123\n')
    testing_pkg.add_file('manual_tests/001.in').write_text('456\n')
    await generate_testcases()

    assert (
        testing_pkg.get_build_testgroup_path('samples') / '000.in'
    ).read_text() == '123\n'
    assert (
        testing_pkg.get_build_testgroup_path('samples') / '001.in'
    ).read_text() == '456\n'
    assert (
        testing_pkg.get_build_testgroup_path('main') / '000.in'
    ).read_text() == '424242\n'


async def test_generator_copy_output_over(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.add_testgroup_from_glob('samples', 'manual_tests/*.in')
    testing_pkg.add_file('manual_tests/000.in').write_text('123\n')
    testing_pkg.add_file('manual_tests/001.in').write_text('456\n')
    await generate_testcases()

    assert (
        testing_pkg.get_build_testgroup_path('samples') / '000.in'
    ).read_text() == '123\n'
    assert (
        testing_pkg.get_build_testgroup_path('samples') / '001.in'
    ).read_text() == '456\n'


async def test_generator_fix_crlf(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.add_testgroup_from_glob('samples', 'manual_tests/*.in')
    testing_pkg.add_file('manual_tests/000.in').write_bytes(b'123\r\n\r\n456\r\n')
    testing_pkg.add_file('manual_tests/001.in').write_bytes(b'456\r\n')
    await generate_testcases()

    assert (
        testing_pkg.get_build_testgroup_path('samples') / '000.in'
    ).read_bytes() == b'123\n\n456\n'
    assert (
        testing_pkg.get_build_testgroup_path('samples') / '001.in'
    ).read_bytes() == b'456\n'


async def test_generator_erases_previously_built_testcases(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_plan('main', 'gens/gen.cpp 123')
    testing_pkg.add_testgroup_from_plan('secondary', 'gens/gen.cpp 456')
    await generate_testcases()

    assert (
        testing_pkg.get_build_testgroup_path('main') / '000.in'
    ).read_bytes() == b'123\n'
    assert (
        testing_pkg.get_build_testgroup_path('secondary') / '000.in'
    ).read_bytes() == b'456\n'

    # Regenerate but only with main group
    await generate_testcases(groups={'main'})

    # Ensure secondary group is not built
    assert (
        testing_pkg.get_build_testgroup_path('main') / '000.in'
    ).read_bytes() == b'123\n'
    assert not (testing_pkg.get_build_testgroup_path('secondary') / '000.in').exists()


async def test_generator_non_existent(
    testing_pkg: testing_package.TestingPackage,
    capsys: pytest.CaptureFixture[str],
):
    testing_pkg.add_testgroup_from_plan('main', 'gens/gen_non_existent.cpp 123\n')

    with pytest.raises(typer.Exit):
        await generate_testcases()

    out = capsys.readouterr().out
    assert 'Generator gens/gen_non_existent.cpp is not present in the package' in out
    assert (
        f'This generator is referenced from {testing_pkg.abspath("testplan/main.txt")}:1'
        in out
    )


async def test_generator_not_compile(
    testing_pkg: testing_package.TestingPackage,
    capsys: pytest.CaptureFixture[str],
):
    testing_pkg.add_generator('gens/gen.cpp').write_text(
        'int main() { cout << 3 << endl; }'
    )
    testing_pkg.add_testgroup_from_plan('main', 'gens/gen.cpp 123')

    with pytest.raises(steps.CompilationError):
        await generate_testcases()

    out = capsys.readouterr().out
    assert 'Failed compiling generator gens/gen.cpp' in out


async def test_generate_standalone_copied_from(
    testing_pkg: testing_package.TestingPackage,
):
    input_file = testing_pkg.add_file('manual_tests/000.in')
    input_file.write_text('123\n')

    tmpd = testing_pkg.mkdtemp()
    spec = GenerationMetadata(
        copied_to=Testcase(inputPath=tmpd / '000.in'),
        copied_from=Testcase(
            inputPath=input_file,
        ),
    )
    await generate_standalone(spec)

    assert (tmpd / '000.in').read_bytes() == b'123\n'


async def test_generate_standalone_generator_call(
    testing_pkg: testing_package.TestingPackage,
):
    input_file = testing_pkg.add_file('manual_tests/000.in')
    input_file.write_text('123\n')

    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')

    tmpd = testing_pkg.mkdtemp()
    spec = GenerationMetadata(
        generator_call=GeneratorCall(
            name='gens/gen.cpp',
            args='456',
        ),
        copied_to=Testcase(inputPath=tmpd / '000.in'),
    )
    await generate_standalone(spec)

    assert (tmpd / '000.in').read_bytes() == b'456\n'


async def test_generate_standalone_validation_works(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.set_validator('validator.cpp', src='validators/int-validator.cpp')

    input_file = testing_pkg.add_file('manual_tests/000.in')
    input_file.write_text('123\n')

    tmpd = testing_pkg.mkdtemp()
    spec = GenerationMetadata(
        copied_to=Testcase(inputPath=tmpd / '000.in'),
        copied_from=Testcase(
            inputPath=input_file,
        ),
    )
    await generate_standalone(spec)

    assert (tmpd / '000.in').read_bytes() == b'123\n'


async def test_generate_standalone_validation_fails(
    testing_pkg: testing_package.TestingPackage,
):
    testing_pkg.set_validator('validator.cpp', src='validators/int-validator.cpp')

    input_file = testing_pkg.add_file('manual_tests/000.in')
    input_file.write_text('123 456\n')

    tmpd = testing_pkg.mkdtemp()
    spec = GenerationMetadata(
        copied_to=Testcase(inputPath=tmpd / '000.in'),
        copied_from=Testcase(
            inputPath=input_file,
        ),
    )

    with pytest.raises(ValidationError) as e:
        await generate_standalone(spec)

        assert 'failed validating testcase' in str(e)


async def test_generate_standalone_package_extra_validator_passes(
    testing_pkg: testing_package.TestingPackage,
):
    # Add a package-level extra validator that requires odd integers
    testing_pkg.add_from_testdata(
        'extra-validator-odd.cpp', src='validators/extra-validator-odd.cpp'
    )
    testing_pkg.yml.extraValidators = testing_pkg.yml.extraValidators + [
        CodeItem(path=pathlib.Path('extra-validator-odd.cpp'))
    ]
    testing_pkg.save()

    # Input satisfies odd constraint
    input_file = testing_pkg.add_file('manual_tests/000.in')
    input_file.write_text('123\n')

    tmpd = testing_pkg.mkdtemp()
    spec = GenerationMetadata(
        copied_to=Testcase(inputPath=tmpd / '000.in'),
        copied_from=Testcase(
            inputPath=input_file,
        ),
    )
    await generate_standalone(spec)

    assert (tmpd / '000.in').read_bytes() == b'123\n'


async def test_generate_standalone_package_extra_validator_fails(
    testing_pkg: testing_package.TestingPackage,
):
    # Add a package-level extra validator that requires odd integers
    testing_pkg.add_from_testdata(
        'extra-validator-odd.cpp', src='validators/extra-validator-odd.cpp'
    )
    testing_pkg.yml.extraValidators = testing_pkg.yml.extraValidators + [
        CodeItem(path=pathlib.Path('extra-validator-odd.cpp'))
    ]
    testing_pkg.save()

    # Input violates odd constraint
    input_file = testing_pkg.add_file('manual_tests/000.in')
    input_file.write_text('100\n')

    tmpd = testing_pkg.mkdtemp()
    spec = GenerationMetadata(
        copied_to=Testcase(inputPath=tmpd / '000.in'),
        copied_from=Testcase(
            inputPath=input_file,
        ),
    )

    with pytest.raises(ValidationError):
        await generate_standalone(spec)


async def test_generate_standalone_reuses_validators_digests_cache(
    testing_pkg: testing_package.TestingPackage, monkeypatch: pytest.MonkeyPatch
):
    # Configure main and extra validators
    testing_pkg.set_validator('validator.cpp', src='validators/int-validator.cpp')
    testing_pkg.add_from_testdata(
        'extra-validator-odd.cpp', src='validators/extra-validator-odd.cpp'
    )
    testing_pkg.yml.extraValidators = testing_pkg.yml.extraValidators + [
        CodeItem(path=pathlib.Path('extra-validator-odd.cpp'))
    ]
    testing_pkg.save()

    # Precompile digests before monkeypatching
    all_validators = validators_mod.compile_validators(
        validators=[
            CodeItem(path=pathlib.Path('validator.cpp')),
            CodeItem(path=pathlib.Path('extra-validator-odd.cpp')),
        ]
    )

    calls = []

    def fake_compile_validators(validators, progress=None):
        paths = [str(v.path) for v in validators]
        calls.append(paths)
        return {p: all_validators[p] for p in paths}

    monkeypatch.setattr(
        validators_mod, 'compile_validators', fake_compile_validators, raising=True
    )

    # Input satisfies all validators
    input_file = testing_pkg.add_file('manual_tests/000.in')
    input_file.write_text('123\n')

    tmpd = testing_pkg.mkdtemp()
    spec = GenerationMetadata(
        copied_to=Testcase(inputPath=tmpd / '000.in'),
        copied_from=Testcase(
            inputPath=input_file,
        ),
    )

    # 1) Provide only main validator digest -> extra should be compiled once
    await generate_standalone(
        spec,
        validators_digests={'validator.cpp': all_validators['validator.cpp']},
    )
    assert calls == [['extra-validator-odd.cpp']]

    # 2) Provide both digests -> no compilation should happen
    calls.clear()
    await generate_standalone(
        spec,
        validators_digests=all_validators,
    )
    assert calls == []


async def test_generator_hash_duplicate_warning(
    testing_pkg: testing_package.TestingPackage,
    capsys: pytest.CaptureFixture[str],
):
    # This generator ignores arguments and always outputs "123"
    testing_pkg.add_generator('gens/gen.cpp').write_text(
        '#include <bits/stdc++.h>\nusing namespace std;\nint main() { cout << 123 << endl; }'
    )
    testing_pkg.add_testgroup_from_plan(
        'main',
        """
gens/gen.cpp 1
gens/gen.cpp 2
""",
    )
    await generate_testcases()

    from rbx.utils import strip_ansi_codes

    out = strip_ansi_codes(capsys.readouterr().out)
    assert (
        'Test main/1 (gens/gen.cpp 2) is a hash duplicate of main/0 (gens/gen.cpp 1).'
        in out
    )


async def test_generator_no_hash_duplicate_warning(
    testing_pkg: testing_package.TestingPackage,
    capsys: pytest.CaptureFixture[str],
):
    # This generator outputs its argument
    testing_pkg.add_generator('gens/gen.cpp', src='generators/gen-id.cpp')
    testing_pkg.add_testgroup_from_plan(
        'main',
        """
gens/gen.cpp 1
gens/gen.cpp 2
""",
    )
    await generate_testcases()

    out = capsys.readouterr().out
    assert 'is a hash duplicate of' not in out



import dataclasses
import pathlib
import re
import shutil
import typing
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Literal, Optional, Tuple, TypeVar

import pypandoc
import typer
from pydantic import BaseModel, Field

from rbx import console, utils
from rbx.box import naming
from rbx.box.fields import Primitive
from rbx.box.schema import LimitsProfile, Package
from rbx.box.statements import texsoup_utils
from rbx.box.statements.demacro_utils import collect_macro_definitions
from rbx.box.statements.latex_jinja import (
    JinjaDictGetter,
    JinjaDictWrapper,
    render_latex_template,
    render_latex_template_blocks,
    render_markdown_template_blocks,
)
from rbx.box.statements.schema import (
    ConversionStep,
    ConversionType,
    JinjaTeX,
    Statement,
    StatementType,
    TexToPDF,
    rbxMarkdownToTeX,
    rbxToTeX,
)
from rbx.box.testcase_sample_utils import StatementSample


@dataclasses.dataclass
class StatementCodeLanguage:
    id: str
    name: str
    command: str


@dataclasses.dataclass
class StatementBuilderContext:
    lang: str
    languages: List[StatementCodeLanguage]
    params: ConversionStep
    root: pathlib.Path
    contest: Optional['StatementBuilderContest'] = None

    def build_jinja_kwargs(self) -> Dict[str, Any]:
        res = {
            'lang': self.lang,
            'languages': self.languages,
            'keyed_languages': {lang.id: lang for lang in self.languages},
        }
        if self.contest is not None:
            contest_kwargs = self.contest.build_jinja_kwargs()
            if 'contest' in contest_kwargs:
                res['contest'] = contest_kwargs['contest']
        return res


class StatementBuilderItem(ABC):
    @abstractmethod
    def build_jinja_kwargs(
        self,
    ) -> Dict[str, Any]:
        pass


class ExplainedStatementSample(StatementSample):
    explanation: Optional[str] = None

    @staticmethod
    def from_statement_sample(
        statement_sample: StatementSample, explanation_block: Optional[str] = None
    ) -> 'ExplainedStatementSample':
        return ExplainedStatementSample(
            **statement_sample.model_dump(),
            explanation=statement_sample.explanationPath.read_text()
            if statement_sample.explanationPath is not None
            and statement_sample.explanationPath.is_file()
            else explanation_block,
        )

    @staticmethod
    def from_statement_samples(
        statement_samples: List[StatementSample],
    ) -> List['ExplainedStatementSample']:
        samples = [
            ExplainedStatementSample.from_statement_sample(
                sample,
            )
            for i, sample in enumerate(statement_samples)
        ]
        return samples


@dataclasses.dataclass
class StatementBuilderProblem(StatementBuilderItem):
    package: Package
    statement: Statement
    limits: LimitsProfile
    profiles: Dict[str, LimitsProfile] = dataclasses.field(default_factory=dict)
    samples: List[StatementSample] = dataclasses.field(default_factory=list)
    short_name: Optional[str] = None

    # Will only be filled by contests.
    io_path: Optional[pathlib.Path] = None

    vars: Optional[Dict[str, Primitive]] = None

    def build_inner_jinja_kwargs(
        self,
    ) -> Dict[str, Any]:
        kwargs = dict(JinjaDictWrapper.from_dict(self.vars or {}, wrapper_key='vars'))
        kwargs.update(
            {
                'package': self.package,
                'statement': self.statement,
                'samples': ExplainedStatementSample.from_statement_samples(
                    self.samples,
                ),
                'vars': JinjaDictWrapper.from_dict(self.vars or {}, wrapper_key='vars'),
                'title': naming.get_problem_title(
                    self.statement.language, self.statement, self.package
                ),
                'limits': self.limits,
                'profiles': JinjaDictGetter('profiles', **self.profiles),
            }
        )
        if self.short_name is not None:
            kwargs['short_name'] = self.short_name
        if self.io_path is not None:
            kwargs['path'] = self.io_path
        return kwargs

    def build_jinja_kwargs(
        self,
    ) -> Dict[str, Any]:
        inner = self.build_inner_jinja_kwargs()
        return {
            'problem': inner,
        }


@dataclasses.dataclass
class StatementBuilderContest(StatementBuilderItem):
    title: str
    location: Optional[str] = None
    date: Optional[str] = None
    problems: List[StatementBuilderProblem] = dataclasses.field(default_factory=list)
    vars: Optional[Dict[str, Primitive]] = None

    def _get_vars(self) -> Dict[str, Primitive]:
        return JinjaDictWrapper.from_dict(self.vars or {}, wrapper_key='vars')

    def build_inner_jinja_kwargs(self) -> Dict[str, Any]:
        res = {'title': self.title, 'vars': self._get_vars()}
        if self.location:
            res['location'] = self.location
        if self.date:
            res['date'] = self.date
        return res

    def build_jinja_kwargs(
        self,
    ) -> Dict[str, Any]:
        res = {
            'contest': self.build_inner_jinja_kwargs(),
            'problems': [
                problem.build_inner_jinja_kwargs() for problem in self.problems
            ],
            'vars': self._get_vars(),  # Kept for backward compatibility.
        }
        return res


class StatementBlocks(BaseModel):
    blocks: Dict[str, str] = Field(default_factory=dict)
    explanations: Dict[int, str] = Field(default_factory=dict)


def prepare_assets(
    assets: List[Tuple[pathlib.Path, pathlib.Path]],
    dest_dir: pathlib.Path,
):
    dest_dir.mkdir(parents=True, exist_ok=True)

    for asset_in, asset_out in assets:
        if not asset_in.is_file():
            console.console.print(
                f'[error]Asset [item]{asset_in}[/item] does not exist in your package.[/error]'
            )
            raise typer.Exit(1)

        # dest_path = dest_dir / asset.resolve().relative_to(statement_dir)
        dest_path = dest_dir / asset_out
        dest_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(str(asset_in), str(dest_path))


def render_jinja(root: pathlib.Path, content: bytes, **kwargs) -> bytes:
    temp_file = '__input__.tex'
    temp_path = root / temp_file
    temp_path.write_bytes(content)

    result: str = render_latex_template(
        str(root),
        temp_file,
        kwargs,
    )
    return result.encode()


def render_jinja_blocks(
    root: pathlib.Path,
    content: bytes,
    mode: Literal['latex', 'markdown'] = 'latex',
    **kwargs,
) -> StatementBlocks:
    if mode == 'latex':
        temp_file = '__input__.tex'
        renderer = render_latex_template_blocks
    elif mode == 'markdown':
        temp_file = '__input__.md'
        renderer = render_markdown_template_blocks
    else:
        raise ValueError(f'Invalid mode: {mode}')

    temp_path = root / temp_file
    temp_path.write_bytes(content)

    result: Dict[str, str] = renderer(
        str(root),
        temp_file,
        kwargs,
    )

    pattern = re.compile(r'explanation_(\d+)')
    explanation_keys = []
    for key in result:
        if match := pattern.match(key):
            explanation_keys.append((key, int(match.group(1))))

    explanations = {value: result[key] for key, value in explanation_keys}
    return StatementBlocks(blocks=result, explanations=explanations)


def _inject_explanations_back(
    statement_blocks: StatementBlocks,
    explained_samples: List[ExplainedStatementSample],
):
    statement_blocks.explanations = {
        i: sample.explanation
        for i, sample in enumerate(explained_samples)
        if sample.explanation is not None
    }


def get_rbxtex_blocks(
    input: bytes,
    context: StatementBuilderContext,
    item: StatementBuilderItem,
    mode: Literal['latex', 'markdown'] = 'latex',
    externalize: bool = False,
) -> Tuple[StatementBlocks, Dict[str, Any]]:
    if isinstance(item, StatementBuilderProblem):
        statement_blocks = render_jinja_blocks(
            context.root,
            input,
            mode=mode,
            **item.build_inner_jinja_kwargs(),
        )
    else:
        statement_blocks = render_jinja_blocks(
            context.root,
            input,
            mode=mode,
            **item.build_jinja_kwargs(),
        )
    if externalize:
        statement_blocks.blocks = externalize_blocks(statement_blocks.blocks)

    item_kwargs = item.build_jinja_kwargs()
    if isinstance(item, StatementBuilderProblem):
        # Build samples.
        for i, sample in enumerate(item_kwargs['problem']['samples']):
            if i in statement_blocks.explanations:
                # Sample will come from a block, not from the file.
                sample.explanation = statement_blocks.explanations[i]
                continue
            if sample.explanation is None:
                # No explanation provided.
                continue
            # Render samples.
            sample.explanation = render_jinja(
                context.root,
                sample.explanation.encode(),
                mode=mode,
                **item.build_inner_jinja_kwargs(),
            ).decode()

        # Externalize samples.
        if externalize:
            item_kwargs['problem']['samples'] = externalize_explained_samples(
                item_kwargs['problem']['samples']
            )

        _inject_explanations_back(statement_blocks, item_kwargs['problem']['samples'])
        item_kwargs['problem']['blocks'] = statement_blocks.blocks
    elif isinstance(item, StatementBuilderContest):
        item_kwargs['contest']['blocks'] = statement_blocks.blocks
    return statement_blocks, item_kwargs


VarBlock = TypeVar('VarBlock')


def externalize_blocks(blocks: Dict[VarBlock, str]) -> Dict[VarBlock, str]:
    res = {}
    for key in blocks:
        tex_node = texsoup_utils.parse_latex(blocks[key])
        texsoup_utils.add_labels_to_tikz_nodes(tex_node, prefix=str(key))
        res[key] = str(tex_node)
    return res


def externalize_explained_samples(
    samples: List[ExplainedStatementSample],
) -> List[ExplainedStatementSample]:
    for idx, sample in enumerate(samples):
        if sample.explanation is None:
            continue
        tex_node = texsoup_utils.parse_latex(sample.explanation)
        texsoup_utils.add_labels_to_tikz_nodes(tex_node, prefix=f'explanation_{idx}')
        sample.explanation = str(tex_node)
    return samples


def substitute_externalized_blocks(blocks: Dict[Any, str]) -> Dict[Any, str]:
    res = {}
    for key in blocks:
        tex_node = texsoup_utils.parse_latex(blocks[key])
        texsoup_utils.replace_labeled_tikz_nodes(tex_node)
        res[key] = str(tex_node)
    return res


def substitute_externalized_samples(
    samples: List[ExplainedStatementSample],
) -> List[ExplainedStatementSample]:
    for sample in samples:
        if sample.explanation is None:
            continue
        tex_node = texsoup_utils.parse_latex(sample.explanation)
        texsoup_utils.replace_labeled_tikz_nodes(tex_node)
        sample.explanation = str(tex_node)
    return samples


class StatementBuilder(ABC):
    @classmethod
    def name(cls) -> ConversionType:
        raise NotImplementedError

    @abstractmethod
    def default_params(self) -> ConversionStep:
        pass

    @abstractmethod
    def input_type(self) -> StatementType:
        pass

    @abstractmethod
    def output_type(self) -> StatementType:
        pass

    def handles_contest(self) -> bool:
        return True

    def handles_problem(self) -> bool:
        return True

    def inject_assets(
        self, root: pathlib.Path, params: ConversionStep
    ) -> List[Tuple[pathlib.Path, pathlib.Path]]:
        return []

    def explanation_suffix(self) -> Optional[str]:
        return None

    @abstractmethod
    def build(
        self,
        input: bytes,
        context: StatementBuilderContext,
        item: StatementBuilderItem,
        verbose: bool = False,
    ) -> bytes:
        pass


class JinjaTeXBuilder(StatementBuilder):
    @classmethod
    def name(cls) -> ConversionType:
        return ConversionType.JinjaTeX

    def default_params(self) -> ConversionStep:
        return JinjaTeX(type=ConversionType.JinjaTeX)

    def input_type(self) -> StatementType:
        return StatementType.JinjaTeX

    def output_type(self) -> StatementType:
        return StatementType.TeX

    def explanation_suffix(self) -> Optional[str]:
        return '.tex'

    def build(
        self,
        input: bytes,
        context: StatementBuilderContext,
        item: StatementBuilderItem,
        verbose: bool = False,
    ) -> bytes:
        return render_jinja(
            context.root,
            input,
            **context.build_jinja_kwargs(),
            **item.build_jinja_kwargs(),
        )


class rbxTeXBuilder(StatementBuilder):
    @classmethod
    def name(cls) -> ConversionType:
        return ConversionType.rbxToTex

    def default_params(self) -> ConversionStep:
        return rbxToTeX(type=ConversionType.rbxToTex)

    def input_type(self) -> StatementType:
        return StatementType.rbxTeX

    def output_type(self) -> StatementType:
        return StatementType.TeX

    def handles_contest(self) -> bool:
        return True

    def inject_assets(
        self, root: pathlib.Path, params: ConversionStep
    ) -> List[Tuple[pathlib.Path, pathlib.Path]]:
        params = typing.cast(rbxToTeX, params)
        if not params.template:
            return []
        return [(utils.abspath(root / params.template), params.template)]

    def explanation_suffix(self) -> Optional[str]:
        return '.tex'

    def build(
        self,
        input: bytes,
        context: StatementBuilderContext,
        item: StatementBuilderItem,
        verbose: bool = False,
    ) -> bytes:
        params = typing.cast(rbxToTeX, context.params)
        assert params.template is not None

        # Get non-externalized version first.
        statement_blocks, item_kwargs = get_rbxtex_blocks(
            input,
            context,
            item,
            mode='latex',
        )
        non_externalized_tex = render_jinja(
            context.root,
            f'%- extends "{params.template}"'.encode(),
            **context.build_jinja_kwargs(),
            **item_kwargs,
            blocks=statement_blocks.blocks,
        )
        (context.root / 'blocks.yml').write_text(utils.model_to_yaml(statement_blocks))
        if not params.externalize:
            return non_externalized_tex

        # Produce externalized version.
        statement_blocks, item_kwargs = get_rbxtex_blocks(
            input,
            context,
            item,
            mode='latex',
            externalize=params.externalize,
        )

        externalized_tex = render_jinja(
            context.root,
            f'%- extends "{params.template}"'.encode(),
            **context.build_jinja_kwargs(),
            **item_kwargs,
            blocks=statement_blocks.blocks,
        )
        (context.root / 'blocks.ext.yml').write_text(
            utils.model_to_yaml(statement_blocks)
        )

        # Produce substituted version.
        statement_blocks.blocks = substitute_externalized_blocks(
            statement_blocks.blocks
        )
        statement_blocks.explanations = substitute_externalized_blocks(
            statement_blocks.explanations
        )
        (context.root / 'blocks.sub.yml').write_text(
            utils.model_to_yaml(statement_blocks)
        )
        return externalized_tex


class rbxMarkdownToTeXBuilder(StatementBuilder):
    @classmethod
    def name(cls) -> ConversionType:
        return ConversionType.rbxMarkdownToTeX

    def default_params(self) -> ConversionStep:
        return rbxMarkdownToTeX(type=ConversionType.rbxMarkdownToTeX)

    def input_type(self) -> StatementType:
        return StatementType.rbxMarkdown

    def output_type(self) -> StatementType:
        return StatementType.rbxTeX

    def handles_contest(self) -> bool:
        return True

    def explanation_suffix(self) -> Optional[str]:
        return '.md'

    def build(
        self,
        input: bytes,
        context: StatementBuilderContext,
        item: StatementBuilderItem,
        verbose: bool = False,
    ) -> bytes:
        statement_blocks, _ = get_rbxtex_blocks(input, context, item, mode='markdown')

        result_str = ''
        for name, content in statement_blocks.blocks.items():
            converted_content = pypandoc.convert_text(content, 'latex', 'markdown')
            result_str += f'%- block {name}\n{converted_content}\n%- endblock\n\n'

        return result_str.encode()


class TeX2PDFBuilder(StatementBuilder):
    @classmethod
    def name(cls) -> ConversionType:
        return ConversionType.TexToPDF

    def default_params(self) -> ConversionStep:
        return TexToPDF(type=ConversionType.TexToPDF)

    def input_type(self) -> StatementType:
        return StatementType.TeX

    def output_type(self) -> StatementType:
        return StatementType.PDF

    def explanation_suffix(self) -> Optional[str]:
        return '.tex'

    def build(
        self,
        input: bytes,
        context: StatementBuilderContext,
        item: StatementBuilderItem,
        verbose: bool = False,
    ) -> bytes:
        from rbx.box.statements.latex import (
            MAX_PDFLATEX_RUNS,
            Latex,
            decode_latex_output,
            should_rerun,
        )

        input_str = input.decode()

        params = typing.cast(TexToPDF, context.params)

        if params.externalize:
            tex_node = texsoup_utils.parse_latex(input_str)
            texsoup_utils.inject_externalization_for_tikz(tex_node)
            (context.root / texsoup_utils.EXTERNALIZATION_DIR).mkdir(
                exist_ok=True, parents=True
            )
            input_str = str(tex_node)

        latex = Latex(input_str)
        latex_result = latex.build_pdf(context.root)
        pdf = latex_result.pdf
        logs = decode_latex_output(latex_result.result.stdout)
        runs = 1

        while pdf is not None and should_rerun(logs) and runs < MAX_PDFLATEX_RUNS:
            console.console.print(
                'Re-running pdfLaTeX to get cross-references right...'
            )
            latex_result = latex.build_pdf(context.root)
            pdf = latex_result.pdf
            logs = decode_latex_output(latex_result.result.stdout)
            runs += 1

        if pdf is None:
            console.console.print(f'{logs}')
            console.console.print('[error]PdfLaTeX compilation failed.[/error]')
            raise typer.Exit(1)

        if verbose:
            console.console.print(f'{logs}')

        if params.demacro:
            macro_defs = collect_macro_definitions(context.root / 'statement.tex')
            macro_defs.to_json_file(context.root / 'macros.json')

        return pdf


BUILDER_LIST: List[StatementBuilder] = [
    TeX2PDFBuilder(),
    JinjaTeXBuilder(),
    rbxTeXBuilder(),
    rbxMarkdownToTeXBuilder(),
]
PROBLEM_BUILDER_LIST = [
    builder for builder in BUILDER_LIST if builder.handles_problem()
]
CONTEST_BUILDER_LIST = [
    builder for builder in BUILDER_LIST if builder.handles_contest()
]



import pathlib
from typing import Any, List, Optional, Union

import typer
from pydantic import BaseModel

from rbx import console, utils
from rbx.box.schema import CodeItem, GeneratorCall, Solution, Testcase, Visualizer
from rbx.box.testcase_utils import TestcaseEntry

TestcaseOrScriptEntry = Union[TestcaseEntry, 'GeneratorScriptEntry']


class GeneratorScriptEntry(BaseModel):
    path: pathlib.Path
    line: int

    def __str__(self) -> str:
        return f'{self.path}:{self.line}'

    def __hash__(self) -> int:
        return hash((self.path, self.line))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, GeneratorScriptEntry):
            return False
        try:
            return self.path.samefile(other.path) and self.line == other.line
        except FileNotFoundError:
            return (
                self.path.absolute() == other.path.absolute()
                and self.line == other.line
            )

    @classmethod
    def parse(cls, spec: str) -> 'GeneratorScriptEntry':
        if spec.count(':') != 1:
            raise ValueError(f'Invalid generator script spec: {spec}')
        path, line = spec.split(':')
        return GeneratorScriptEntry(path=pathlib.Path(path), line=int(line))


class GenerationInput(BaseModel):
    copied_from: Optional[Testcase] = None
    generator_call: Optional[GeneratorCall] = None
    generator_script: Optional[GeneratorScriptEntry] = None
    content: Optional[str] = None


class GenerationMetadata(GenerationInput):
    copied_to: Testcase

    def __str__(self) -> str:
        if self.generator_call is not None:
            return utils.escape_markup(str(self.generator_call))
        elif self.copied_from is not None:
            return f'{self.copied_from.inputPath}'
        return ''


class GenerationTestcaseEntry(BaseModel):
    group_entry: TestcaseEntry
    subgroup_entry: TestcaseEntry

    metadata: GenerationMetadata
    validator: Optional[CodeItem] = None
    extra_validators: List[CodeItem] = []
    output_validators: List[CodeItem] = []
    model_solution: Optional[Solution] = None

    visualizer: Optional[Visualizer] = None
    solution_visualizer: Optional[Visualizer] = None

    def is_sample(self) -> bool:
        return self.group_entry.group == 'samples'

    def __str__(self) -> str:
        result = f'{self.group_entry}'
        metadata_str = str(self.metadata)
        if metadata_str:
            result += f' ({metadata_str})'
        return result


def get_parsed_entry(spec: str) -> TestcaseOrScriptEntry:
    try:
        if spec.count(':') == 1:
            return GeneratorScriptEntry.parse(spec)
        elif spec.count('/') == 1:
            return TestcaseEntry.parse(spec)
    except Exception:
        pass

    console.console.print(f'[error]Invalid testcase spec: {spec}[/error]')
    raise typer.Exit(1)
